from __future__ import annotations

from typing import TYPE_CHECKING, Any

from django.conf import settings
from django.http.request import MediaType

from undine.dataclasses import GraphQLHttpParams
from undine.exceptions import (
    GraphQLMissingContentTypeError,
    GraphQLMissingDocumentIDError,
    GraphQLMissingFileMapError,
    GraphQLMissingOperationsError,
    GraphQLMissingQueryAndDocumentIDError,
    GraphQLMissingQueryError,
    GraphQLPersistedDocumentNotFoundError,
    GraphQLPersistedDocumentsNotSupportedError,
    GraphQLRequestDecodingError,
    GraphQLUnsupportedContentTypeError,
)
from undine.http.files import place_files
from undine.http.utils import decode_body, load_json_dict, parse_json_body
from undine.settings import undine_settings
from undine.utils.reflection import is_list_of

if TYPE_CHECKING:
    from django.core.files.uploadedfile import UploadedFile

    from undine.typing import DjangoRequestProtocol

__all__ = [
    "GraphQLRequestParamsParser",
]


class GraphQLRequestParamsParser:
    """Parse GraphQLParams from a given HttpRequest."""

    @classmethod
    def run(cls, request: DjangoRequestProtocol) -> GraphQLHttpParams:
        data = cls.parse_body(request)
        return cls.get_graphql_params(data)

    @classmethod
    def parse_body(cls, request: DjangoRequestProtocol) -> dict[str, str]:
        if request.method == "GET":
            return request.GET.dict()  # type: ignore[return-value]

        if not request.content_type:
            raise GraphQLMissingContentTypeError

        content_type = MediaType(request.content_type)
        charset = str(content_type.params.get("charset", "utf-8"))

        if content_type.main_type == "application":
            if content_type.sub_type == "json":
                return parse_json_body(request.body, charset=charset)
            if content_type.sub_type == "graphql":
                return {"query": decode_body(request.body, charset=charset)}
            if content_type.sub_type == "x-www-form-urlencoded":
                return request.POST.dict()  # type: ignore[return-value]

        if (
            undine_settings.FILE_UPLOAD_ENABLED
            and content_type.main_type == "multipart"
            and content_type.sub_type == "form-data"
        ):
            if request.FILES:
                return cls.parse_file_uploads(request.POST.dict(), request.FILES.dict())  # type: ignore[arg-type]
            return request.POST.dict()  # type: ignore[return-value]

        raise GraphQLUnsupportedContentTypeError(content_type=content_type)

    @classmethod
    def parse_file_uploads(cls, post_data: dict[str, str], files: dict[str, UploadedFile]) -> dict[str, Any]:
        operations = cls.get_operations(post_data)
        files_map = cls.get_map(post_data)
        place_files(operations, files_map, files)
        return operations

    @classmethod
    def get_operations(cls, post_data: dict[str, str]) -> dict[str, Any]:
        operations: str | None = post_data.get("operations")
        if not isinstance(operations, str):
            raise GraphQLMissingOperationsError

        return load_json_dict(
            operations,
            decode_error_msg="The `operations` value must be a JSON object.",
            type_error_msg="The `operations` value is not a mapping.",
        )

    @classmethod
    def get_map(cls, post_data: dict[str, str]) -> dict[str, list[str]]:
        files_map_str: str | None = post_data.get("map")
        if not isinstance(files_map_str, str):
            raise GraphQLMissingFileMapError

        files_map = load_json_dict(
            files_map_str,
            decode_error_msg="The `map` value must be a JSON object.",
            type_error_msg="The `map` value is not a mapping.",
        )

        for value in files_map.values():
            if not is_list_of(value, str, allow_empty=True):
                msg = "The `map` value is not a mapping from string to list of strings."
                raise GraphQLRequestDecodingError(msg)

        return files_map

    @classmethod
    def get_graphql_params(cls, data: dict[str, Any]) -> GraphQLHttpParams:
        document = cls.parse_document(data)

        operation_name: str | None = data.get("operationName") or None

        variables: dict[str, str] | str | None = data.get("variables")
        if isinstance(variables, str):
            variables = load_json_dict(
                variables,
                decode_error_msg="Variables are invalid JSON.",
                type_error_msg="Variables must be a mapping.",
            )

        extensions: dict[str, str] | str | None = data.get("extensions")
        if isinstance(extensions, str):
            extensions = load_json_dict(
                extensions,
                decode_error_msg="Extensions are invalid JSON.",
                type_error_msg="Extensions must be a mapping.",
            )

        return GraphQLHttpParams(
            document=document,
            variables=variables or {},
            operation_name=operation_name,
            extensions=extensions or {},
        )

    @classmethod
    def parse_document(cls, data: dict[str, Any]) -> str:
        persisted_documents_installed = "undine.persisted_documents" in settings.INSTALLED_APPS

        if not undine_settings.PERSISTED_DOCUMENTS_ONLY:
            query = data.get("query")
            if query:
                return query

            if not persisted_documents_installed:
                raise GraphQLMissingQueryError

        if not persisted_documents_installed:
            raise GraphQLPersistedDocumentsNotSupportedError

        document_id = data.get("documentId")
        if not document_id:
            if undine_settings.PERSISTED_DOCUMENTS_ONLY:
                raise GraphQLMissingDocumentIDError
            raise GraphQLMissingQueryAndDocumentIDError

        return cls.get_persisted_document(document_id)

    @classmethod
    def get_persisted_document(cls, document_id: str) -> str:
        from undine.persisted_documents.models import PersistedDocument  # noqa: PLC0415

        try:
            persisted_document = PersistedDocument.objects.get(document_id=document_id)
        except PersistedDocument.DoesNotExist as error:
            raise GraphQLPersistedDocumentNotFoundError(document_id=document_id) from error

        return persisted_document.document



from __future__ import annotations

import asyncio
import datetime
import types
import uuid
from collections.abc import AsyncGenerator, AsyncIterable, AsyncIterator
from contextlib import suppress
from decimal import Decimal
from enum import Enum, IntEnum, StrEnum
from importlib import import_module
from types import FunctionType
from typing import Any, Union, get_origin, is_typeddict

from django.contrib.contenttypes.fields import GenericForeignKey, GenericRel, GenericRelation
from django.db.models import (
    BinaryField,
    BooleanField,
    CharField,
    DateField,
    DateTimeField,
    DecimalField,
    DurationField,
    EmailField,
    F,
    FileField,
    FloatField,
    ForeignKey,
    GenericIPAddressField,
    ImageField,
    IntegerChoices,
    IntegerField,
    IPAddressField,
    JSONField,
    ManyToManyField,
    ManyToManyRel,
    ManyToOneRel,
    Model,
    OneToOneField,
    OneToOneRel,
    Q,
    TextChoices,
    TextField,
    TimeField,
    URLField,
    UUIDField,
)
from django.db.models.fields.related_descriptors import (
    ForwardManyToOneDescriptor,
    ManyToManyDescriptor,
    ReverseManyToOneDescriptor,
    ReverseOneToOneDescriptor,
)
from django.db.models.query_utils import DeferredAttribute
from django.utils.encoding import force_str
from graphql import (
    GraphQLBoolean,
    GraphQLEnumType,
    GraphQLEnumValue,
    GraphQLField,
    GraphQLFloat,
    GraphQLInputField,
    GraphQLInputObjectType,
    GraphQLInputType,
    GraphQLInt,
    GraphQLInterfaceType,
    GraphQLList,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLOutputType,
    GraphQLScalarType,
    GraphQLString,
    GraphQLUnionType,
    Undefined,
)

from undine import Calculation, Field, InterfaceField, InterfaceType, MutationType, QueryType, UnionType
from undine.converters import convert_lookup_to_graphql_type, convert_to_description, convert_to_graphql_type
from undine.dataclasses import (
    LazyGenericForeignKey,
    LazyLambda,
    LazyRelation,
    LookupRef,
    MaybeManyOrNonNull,
    TypeRef,
    UnionFilterRef,
)
from undine.exceptions import (
    FunctionDispatcherError,
    GraphQLTypedDictAnnotatedIncorrectMetadataError,
    RegistryMissingTypeError,
)
from undine.mutation import Input, MutationTypeMeta
from undine.pagination import OffsetPagination
from undine.parsers import parse_first_param_type, parse_is_nullable, parse_return_annotation
from undine.relay import Connection, PageInfoType
from undine.resolvers.query import NamedTupleFieldResolver, TypedDictFieldResolver
from undine.scalars import (
    GraphQLAny,
    GraphQLBase64,
    GraphQLDate,
    GraphQLDateTime,
    GraphQLDecimal,
    GraphQLDuration,
    GraphQLEmail,
    GraphQLFile,
    GraphQLImage,
    GraphQLIP,
    GraphQLIPv4,
    GraphQLIPv6,
    GraphQLJSON,
    GraphQLTime,
    GraphQLURL,
    GraphQLUUID,
)
from undine.settings import undine_settings
from undine.subscriptions import QueryTypeSignalSubscription
from undine.typing import CombinableExpression, ModelField, eval_type
from undine.utils.graphql.type_registry import (
    get_or_create_graphql_enum,
    get_or_create_graphql_input_object_type,
    get_or_create_graphql_object_type,
)
from undine.utils.model_fields import TextChoicesField
from undine.utils.model_utils import generic_relations_for_generic_foreign_key, get_model_field
from undine.utils.reflection import FunctionEqualityWrapper, get_flattened_generic_params, is_annotated, is_namedtuple
from undine.utils.text import to_camel_case, to_pascal_case, to_schema_name

# --- Python types -------------------------------------------------------------------------------------------------


@convert_to_graphql_type.register
def _(ref: str, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    model_field = get_model_field(model=kwargs["model"], lookup=ref)
    return convert_to_graphql_type(model_field, **kwargs)


@convert_to_graphql_type.register
def _(_: type[str], **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLString


@convert_to_graphql_type.register
def _(_: type[bool], **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLBoolean


@convert_to_graphql_type.register
def _(_: type[int], **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLInt


@convert_to_graphql_type.register
def _(_: type[float], **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLFloat


@convert_to_graphql_type.register
def _(_: type[Decimal], **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLDecimal


@convert_to_graphql_type.register
def _(_: type[datetime.datetime], **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLDateTime


@convert_to_graphql_type.register
def _(_: type[datetime.date], **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLDate


@convert_to_graphql_type.register
def _(_: type[datetime.time], **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLTime


@convert_to_graphql_type.register
def _(_: type[datetime.timedelta], **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLDuration


@convert_to_graphql_type.register
def _(_: type[uuid.UUID], **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLUUID


@convert_to_graphql_type.register
def _(ref: type[Enum | StrEnum], **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return get_or_create_graphql_enum(
        name=ref.__name__,
        values={
            str(value.value): GraphQLEnumValue(value=value, description=str(value.value))
            for name, value in ref.__members__.items()
        },
        description=convert_to_description(ref),
    )


@convert_to_graphql_type.register
def _(ref: type[IntEnum], **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return get_or_create_graphql_enum(
        name=ref.__name__,
        values={
            str(value.name): GraphQLEnumValue(value=value, description=str(value.name))
            for name, value in ref.__members__.items()
        },
        description=convert_to_description(ref),
    )


@convert_to_graphql_type.register
def _(ref: type[TextChoices], **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return get_or_create_graphql_enum(
        name=ref.__name__,
        values={
            str(value.value): GraphQLEnumValue(value=value, description=force_str(value.label))
            for key, value in ref.__members__.items()
        },
        description=convert_to_description(ref),
    )


@convert_to_graphql_type.register
def _(ref: type[IntegerChoices], **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return get_or_create_graphql_enum(
        name=ref.__name__,
        values={
            str(value.name): GraphQLEnumValue(value=value, description=force_str(value.label))
            for key, value in ref.__members__.items()
        },
        description=convert_to_description(ref),
    )


@convert_to_graphql_type.register
def _(_: type, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLAny


@convert_to_graphql_type.register
def _(ref: type[list], **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    args = get_flattened_generic_params(ref)

    nullable = types.NoneType in args
    if nullable:
        args = tuple(arg for arg in args if arg is not types.NoneType)

    # For lists without type, or with a union type, default to any.
    if len(args) != 1:
        return GraphQLList(GraphQLAny)

    graphql_type = convert_to_graphql_type(args[0], **kwargs)
    if not nullable:
        graphql_type = GraphQLNonNull(graphql_type)

    return GraphQLList(graphql_type)


@convert_to_graphql_type.register
def _(ref: type[dict], **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    if not is_typeddict(ref):
        return GraphQLJSON

    module_globals = vars(import_module(ref.__module__))
    is_input = kwargs.get("is_input", False)
    total: bool = getattr(ref, "__total__", True)

    description = convert_to_description(ref)

    if is_input:
        input_fields: dict[str, GraphQLInputField] = {}
        for key, value in ref.__annotations__.items():
            evaluated_type = eval_type(value, globals_=module_globals)

            if is_annotated(evaluated_type):
                inpt = evaluated_type.__metadata__[0]
                if isinstance(inpt, str):
                    inpt = Input(description=inpt)

                if not isinstance(inpt, Input):
                    raise GraphQLTypedDictAnnotatedIncorrectMetadataError(typed_dict=ref, field=key, expected=Input)

                inpt.ref = evaluated_type.__args__[0]

                input_fields[to_schema_name(key)] = GraphQLInputField(
                    convert_to_graphql_type(TypeRef(inpt.ref, total=total), **kwargs),
                    default_value=inpt.default_value,
                    description=inpt.description if inpt.description is not Undefined else None,
                    deprecation_reason=inpt.deprecation_reason,
                    out_name=key,
                    extensions=inpt.extensions,
                )
                continue

            input_fields[to_schema_name(key)] = GraphQLInputField(
                convert_to_graphql_type(TypeRef(evaluated_type, total=total), **kwargs),
                out_name=key,
            )

        return get_or_create_graphql_input_object_type(
            name=ref.__name__.removesuffix("Type").removesuffix("Input") + "Input",
            fields=input_fields,
            description=description,
        )

    output_fields: dict[str, GraphQLField] = {}
    for key, value in ref.__annotations__.items():
        evaluated_type = eval_type(value, globals_=module_globals)
        resolver = TypedDictFieldResolver(key=key)

        if is_annotated(evaluated_type):
            field = evaluated_type.__metadata__[0]

            if isinstance(field, str):
                field = Field(description=field)

            if not isinstance(field, Field):
                raise GraphQLTypedDictAnnotatedIncorrectMetadataError(typed_dict=ref, field=key, expected=Field)

            field.ref = evaluated_type.__args__[0]

            output_fields[to_schema_name(key)] = GraphQLField(
                convert_to_graphql_type(TypeRef(field.ref, total=total), **kwargs),
                resolve=resolver,
                description=field.description if field.description is not Undefined else None,
                deprecation_reason=field.deprecation_reason,
                extensions=field.extensions,
            )
            continue

        output_fields[to_schema_name(key)] = GraphQLField(
            convert_to_graphql_type(TypeRef(evaluated_type, total=total), **kwargs),
            resolve=resolver,
        )

    return get_or_create_graphql_object_type(
        name=ref.__name__.removesuffix("Type") + "Type",
        fields=output_fields,
        description=description,
    )


@convert_to_graphql_type.register
def _(ref: type[tuple], **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    if not is_namedtuple(ref):
        args = get_flattened_generic_params(ref)
        args = tuple(arg for arg in args if arg is not Ellipsis)

        nullable = types.NoneType in args
        if nullable:
            args = tuple(arg for arg in args if arg is not types.NoneType)

        if len(args) != 1:
            return GraphQLList(GraphQLAny)

        graphql_type = convert_to_graphql_type(args[0], **kwargs)
        if not nullable:
            graphql_type = GraphQLNonNull(graphql_type)

        return GraphQLList(graphql_type)

    module_globals = vars(import_module(ref.__module__))
    is_input = kwargs.get("is_input", False)

    description = convert_to_description(ref)

    if is_input:
        input_fields: dict[str, GraphQLInputField] = {}
        defaults: dict[str, Any] = getattr(ref, "_field_defaults", {})

        for key, value in ref.__annotations__.items():
            evaluated_type = eval_type(value, globals_=module_globals)
            input_type: GraphQLInputType = convert_to_graphql_type(TypeRef(evaluated_type), **kwargs)  # type: ignore[assignment]

            input_fields[to_schema_name(key)] = GraphQLInputField(
                input_type,
                default_value=defaults.get(key, Undefined),
                out_name=key,
            )

        return get_or_create_graphql_input_object_type(
            name=ref.__name__.removesuffix("Type").removesuffix("Input") + "Input",
            fields=input_fields,
            description=description,
        )

    output_fields: dict[str, GraphQLField] = {}
    for key, value in ref.__annotations__.items():
        evaluated_type = eval_type(value, globals_=module_globals)
        output_type: GraphQLOutputType = convert_to_graphql_type(TypeRef(evaluated_type), **kwargs)  # type: ignore[assignment]
        resolver = NamedTupleFieldResolver(attr=key)
        output_fields[to_schema_name(key)] = GraphQLField(output_type, resolve=resolver)

    return get_or_create_graphql_object_type(
        name=ref.__name__.removesuffix("Type") + "Type",
        fields=output_fields,
        description=description,
    )


@convert_to_graphql_type.register
def _(ref: FunctionType, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    is_input = kwargs.get("is_input", False)
    annotation = parse_first_param_type(ref) if is_input else parse_return_annotation(ref)
    return convert_to_graphql_type(annotation, **kwargs)


@convert_to_graphql_type.register
def _(ref: type[AsyncGenerator | AsyncIterator | AsyncIterable], **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    if not hasattr(ref, "__args__"):
        msg = f"Cannot convert {ref!r} to GraphQL type without generic type arguments."
        raise FunctionDispatcherError(msg)

    return_type = ref.__args__[0]  # type: ignore[attr-defined]

    origin = get_origin(return_type)
    if origin not in {types.UnionType, Union}:
        return convert_to_graphql_type(TypeRef(return_type), **kwargs)

    args = get_flattened_generic_params(return_type)
    nullable = types.NoneType in args

    # Returning exceptions can be used to emit errors without closing the subscription.
    args = tuple(arg for arg in args if arg is not types.NoneType and not issubclass(arg, BaseException))

    if len(args) != 1:
        return GraphQLAny

    graphql_type = convert_to_graphql_type(TypeRef(args[0]), **kwargs)
    if nullable and isinstance(graphql_type, GraphQLNonNull):
        graphql_type = graphql_type.of_type

    return graphql_type


@convert_to_graphql_type.register
def _(ref: type[asyncio.Future | asyncio.Task], **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    if not hasattr(ref, "__args__"):
        msg = f"Cannot convert {ref!r} to GraphQL type without generic type arguments."
        raise FunctionDispatcherError(msg)

    return_type = ref.__args__[0]  # type: ignore[attr-defined]

    origin = get_origin(return_type)
    if origin not in {types.UnionType, Union}:
        return convert_to_graphql_type(TypeRef(return_type), **kwargs)

    args = get_flattened_generic_params(return_type)
    nullable = types.NoneType in args

    # Exceptions can be returned by DataLoaders in order to raise errors.
    args = tuple(arg for arg in args if arg is not types.NoneType and not issubclass(arg, BaseException))

    if len(args) != 1:
        return GraphQLAny

    graphql_type = convert_to_graphql_type(TypeRef(args[0]), **kwargs)
    if nullable and isinstance(graphql_type, GraphQLNonNull):
        graphql_type = graphql_type.of_type

    return graphql_type


# --- Model fields -------------------------------------------------------------------------------------------------


@convert_to_graphql_type.register
def _(ref: CharField, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    if ref.choices is None:
        return GraphQLString

    # Generate a name for an enum based on the field it is used in.
    # This is required, since CharField doesn't know the name of the enum it is used in.
    # Use `TextChoicesField` instead to not get multiple enums in the GraphQL schema for different fields.
    name = ref.model.__name__ + to_pascal_case(ref.name) + "Choices"

    return get_or_create_graphql_enum(
        name=name,
        values={key: str(value) for key, value in ref.choices},
        description=convert_to_description(ref),
    )


@convert_to_graphql_type.register
def _(_: TextField, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLString


@convert_to_graphql_type.register
def _(ref: TextChoicesField, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return get_or_create_graphql_enum(
        name=ref.choices_enum.__name__,
        values={
            str(value.value): GraphQLEnumValue(value=value, description=str(value.label))
            for key, value in ref.choices_enum.__members__.items()
        },
        description=convert_to_description(ref) or convert_to_description(ref.choices_enum),
    )


@convert_to_graphql_type.register
def _(_: BooleanField, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLBoolean


@convert_to_graphql_type.register
def _(_: IntegerField, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLInt


@convert_to_graphql_type.register
def _(_: FloatField, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLFloat


@convert_to_graphql_type.register
def _(_: DecimalField, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLDecimal


@convert_to_graphql_type.register
def _(_: DateField, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLDate


@convert_to_graphql_type.register
def _(_: DateTimeField, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLDateTime


@convert_to_graphql_type.register
def _(_: TimeField, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLTime


@convert_to_graphql_type.register
def _(_: DurationField, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLDuration


@convert_to_graphql_type.register
def _(_: UUIDField, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLUUID


@convert_to_graphql_type.register
def _(_: EmailField, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLEmail


@convert_to_graphql_type.register
def _(_: IPAddressField, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLIPv4


@convert_to_graphql_type.register
def _(ref: GenericIPAddressField, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    if ref.protocol.lower() == "ipv4":
        return GraphQLIPv4
    if ref.protocol.lower() == "ipv6":
        return GraphQLIPv6
    return GraphQLIP


@convert_to_graphql_type.register
def _(_: URLField, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLURL


@convert_to_graphql_type.register
def _(_: BinaryField, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLBase64


@convert_to_graphql_type.register
def _(_: JSONField, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLJSON


@convert_to_graphql_type.register
def _(_: FileField, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLFile


@convert_to_graphql_type.register
def _(_: ImageField, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLImage


@convert_to_graphql_type.register
def _(ref: OneToOneField, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return convert_to_graphql_type(ref.target_field, **kwargs)


@convert_to_graphql_type.register
def _(ref: ForeignKey, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return convert_to_graphql_type(ref.target_field, **kwargs)


@convert_to_graphql_type.register
def _(ref: ManyToManyField, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    type_ = convert_to_graphql_type(ref.target_field, **kwargs)
    return GraphQLList(GraphQLNonNull(type_))


@convert_to_graphql_type.register
def _(ref: OneToOneRel, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return convert_to_graphql_type(ref.target_field, **kwargs)


@convert_to_graphql_type.register
def _(ref: ManyToOneRel, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    type_ = convert_to_graphql_type(ref.target_field, **kwargs)
    return GraphQLList(GraphQLNonNull(type_))


@convert_to_graphql_type.register
def _(ref: ManyToManyRel, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    type_ = convert_to_graphql_type(ref.target_field, **kwargs)
    return GraphQLList(GraphQLNonNull(type_))


@convert_to_graphql_type.register
def _(ref: GenericForeignKey, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    if not kwargs.get("is_input"):
        field: ModelField = ref.model._meta.get_field(ref.fk_field)
        return convert_to_graphql_type(field, **kwargs)

    fk_field_name = to_pascal_case(ref.name)
    input_object_name = f"{ref.model.__name__}{fk_field_name}Input"

    related_models = [field.model for field in generic_relations_for_generic_foreign_key(ref)]

    def fields() -> dict[str, GraphQLInputField]:
        field_map: dict[str, GraphQLInputField] = {}

        for model in related_models:
            schema_name = f"{ref.model.__name__}{fk_field_name}{model.__name__}Input"

            MutationTypeMeta.__model__ = model

            class RelatedMutation(MutationType, kind="related", schema_name=schema_name): ...

            field_name = to_camel_case(model.__name__)
            input_type = RelatedMutation.__input_type__()

            field_map[field_name] = GraphQLInputField(input_type)

        return field_map

    return get_or_create_graphql_input_object_type(
        name=input_object_name,
        fields=FunctionEqualityWrapper(fields, context=ref),
        is_one_of=True,
    )


@convert_to_graphql_type.register
def _(ref: GenericRelation, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    object_id_field = ref.related_model._meta.get_field(ref.object_id_field_name)
    type_ = convert_to_graphql_type(object_id_field, **kwargs)
    return GraphQLList(type_)


@convert_to_graphql_type.register
def _(ref: GenericRel, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return convert_to_graphql_type(ref.field)


# Postgres fields
with suppress(ImportError):
    from django.contrib.postgres.fields import ArrayField, HStoreField

    @convert_to_graphql_type.register
    def _(_: HStoreField, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
        return GraphQLJSON

    @convert_to_graphql_type.register
    def _(ref: ArrayField, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
        inner_type = convert_to_graphql_type(ref.base_field, **kwargs)
        if not ref.base_field.null:
            inner_type = GraphQLNonNull(inner_type)
        return GraphQLList(inner_type)


# Generated field
with suppress(ImportError):
    from django.db.models import GeneratedField

    @convert_to_graphql_type.register
    def _(ref: GeneratedField, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
        return convert_to_graphql_type(ref.output_field, **kwargs)


# --- Django ORM ---------------------------------------------------------------------------------------------------


@convert_to_graphql_type.register
def _(ref: type[Model], **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return convert_to_graphql_type(ref._meta.pk, **kwargs)


@convert_to_graphql_type.register
def _(ref: F, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    model: type[Model] = kwargs["model"]
    model_field = get_model_field(model=model, lookup=ref.name)
    return convert_to_graphql_type(model_field, **kwargs)


@convert_to_graphql_type.register
def _(_: Q, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLBoolean


@convert_to_graphql_type.register
def _(ref: CombinableExpression, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return convert_to_graphql_type(ref.output_field, **kwargs)


@convert_to_graphql_type.register
def _(ref: DeferredAttribute | ForwardManyToOneDescriptor, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return convert_to_graphql_type(ref.field, **kwargs)


@convert_to_graphql_type.register
def _(ref: ReverseManyToOneDescriptor, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return convert_to_graphql_type(ref.rel, **kwargs)


@convert_to_graphql_type.register
def _(ref: ReverseOneToOneDescriptor, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return convert_to_graphql_type(ref.related, **kwargs)


@convert_to_graphql_type.register
def _(ref: ManyToManyDescriptor, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return convert_to_graphql_type(ref.rel if ref.reverse else ref.field, **kwargs)


# --- GraphQL types ------------------------------------------------------------------------------------------------


@convert_to_graphql_type.register
def _(
    ref: (
        GraphQLScalarType
        | GraphQLObjectType
        | GraphQLInputObjectType
        | GraphQLInterfaceType
        | GraphQLUnionType
        | GraphQLEnumType
        | GraphQLList
        | GraphQLNonNull
    ),
    **kwargs: Any,
) -> GraphQLInputType | GraphQLOutputType:
    return ref


@convert_to_graphql_type.register
def _(ref: GraphQLInterfaceType, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return ref


# --- Custom types -------------------------------------------------------------------------------------------------


@convert_to_graphql_type.register
def _(ref: type[QueryType], **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return ref.__output_type__()


@convert_to_graphql_type.register
def _(ref: type[MutationType], **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    if not kwargs.get("is_input"):
        return ref.__output_type__()

    return ref.__input_type__()


@convert_to_graphql_type.register
def _(ref: type[UnionType], **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return ref.__union_type__()


@convert_to_graphql_type.register
def _(ref: LazyRelation, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    try:
        value = ref.get_type()
    except RegistryMissingTypeError:
        value = ref.field

    return convert_to_graphql_type(value, **kwargs)


@convert_to_graphql_type.register
def _(ref: LazyGenericForeignKey, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    name = ref.field.model.__name__ + to_pascal_case(ref.field.name)

    type(UnionType).__query_types__ = ref.get_types()

    class GenericUnion(UnionType, schema_name=name): ...

    return convert_to_graphql_type(GenericUnion)


@convert_to_graphql_type.register
def _(ref: LazyLambda, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return convert_to_graphql_type(ref.callback(), **kwargs)


@convert_to_graphql_type.register
def _(ref: TypeRef, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    value = convert_to_graphql_type(ref.value, **kwargs)
    nullable = parse_is_nullable(ref.value, is_input=kwargs.get("is_input", False), total=ref.total)
    if not nullable:
        value = GraphQLNonNull(value)
    return value


@convert_to_graphql_type.register
def _(ref: MaybeManyOrNonNull, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    value = convert_to_graphql_type(ref.value, **kwargs)

    # Note that order matters here!

    if ref.nullable is True and isinstance(value, GraphQLNonNull):
        value = value.of_type

    if ref.many is True and (
        not isinstance(value, GraphQLList)
        and not (isinstance(value, GraphQLNonNull) and isinstance(value.of_type, GraphQLList))
    ):
        if not isinstance(value, GraphQLNonNull):
            value = GraphQLNonNull(value)
        value = GraphQLList(value)

    if ref.nullable is False and not isinstance(value, GraphQLNonNull):
        value = GraphQLNonNull(value)

    return value


@convert_to_graphql_type.register
def _(ref: UnionFilterRef, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    kwargs["model"] = ref.models[0]
    return convert_to_graphql_type(ref.ref, **kwargs)


@convert_to_graphql_type.register
def _(ref: type[Calculation], **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return convert_to_graphql_type(TypeRef(ref.__returns__), **kwargs)


@convert_to_graphql_type.register
def _(ref: LookupRef, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    kwargs["default_type"] = convert_to_graphql_type(ref.ref, **kwargs)
    return convert_lookup_to_graphql_type(ref.lookup, **kwargs)


@convert_to_graphql_type.register
def _(ref: Connection, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    if ref.union_type is not None:
        ref_type = ref.union_type
    elif ref.interface_type is not None:
        ref_type = ref.interface_type
    else:
        ref_type = ref.query_type

    def edge_fields() -> dict[str, GraphQLField]:
        return {
            "cursor": GraphQLField(
                GraphQLNonNull(GraphQLString),
                description="A value identifying this edge for pagination purposes.",
            ),
            "node": GraphQLField(
                convert_to_graphql_type(ref_type, **kwargs),  # type: ignore[arg-type]
                description="An item in the connection.",
            ),
        }

    EdgeType = get_or_create_graphql_object_type(  # noqa: N806
        name=f"{ref_type.__schema_name__}Edge",
        description="An object describing an item in the connection.",
        fields=FunctionEqualityWrapper(edge_fields, context=ref),
    )

    return get_or_create_graphql_object_type(
        name=f"{ref_type.__schema_name__}Connection",
        description="A connection to a list of items.",
        fields={
            undine_settings.TOTAL_COUNT_PARAM_NAME: GraphQLField(
                GraphQLNonNull(GraphQLInt),
                description="Total number of items in the connection.",
            ),
            "pageInfo": GraphQLField(
                GraphQLNonNull(PageInfoType),
                description="Information about the current state of the pagination.",
            ),
            "edges": GraphQLField(
                GraphQLList(EdgeType),
                description="The items in the connection.",
            ),
        },
        extensions={
            undine_settings.CONNECTION_EXTENSIONS_KEY: ref,
        },
    )


@convert_to_graphql_type.register
def _(ref: OffsetPagination, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    if ref.union_type is not None:
        return convert_to_graphql_type(ref.union_type, **kwargs)
    if ref.interface_type is not None:
        return convert_to_graphql_type(ref.interface_type, **kwargs)
    return convert_to_graphql_type(ref.query_type, **kwargs)


@convert_to_graphql_type.register
def _(ref: type[InterfaceType], **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return ref.__interface__()


@convert_to_graphql_type.register
def _(ref: InterfaceField, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return ref.output_type


@convert_to_graphql_type.register
def _(ref: QueryTypeSignalSubscription, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return convert_to_graphql_type(ref.query_type, **kwargs)


with suppress(ImportError):
    from undine.utils.full_text_search import PostgresFTS

    @convert_to_graphql_type.register
    def _(_: PostgresFTS, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
        return GraphQLString



from __future__ import annotations

import itertools
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any

from graphql import (
    FieldNode,
    FragmentSpreadNode,
    GraphQLInterfaceType,
    GraphQLResolveInfo,
    GraphQLUnionType,
    InlineFragmentNode,
)
from graphql.execution.collect_fields import get_field_entry_key

from undine.dataclasses import AbstractSelections
from undine.exceptions import GraphQLOptimizerError, ModelFieldError
from undine.settings import undine_settings
from undine.utils.graphql.undine_extensions import get_undine_interface_type, get_undine_query_type
from undine.utils.graphql.utils import (
    get_field_def,
    get_underlying_type,
    is_connection,
    is_edge,
    is_node_interface,
    is_page_info,
    is_relation_id,
    is_typename_metafield,
    should_skip_node,
)
from undine.utils.model_utils import get_model_field, is_generic_foreign_key, is_to_many, is_to_one
from undine.utils.text import to_snake_case

if TYPE_CHECKING:
    from collections.abc import Generator

    from django.contrib.contenttypes.fields import GenericForeignKey
    from django.db.models import Field, Model
    from graphql import (
        FragmentDefinitionNode,
        GraphQLAbstractType,
        GraphQLCompositeType,
        GraphQLNamedOutputType,
        GraphQLObjectType,
        GraphQLScalarType,
    )

    from undine.typing import GQLInfo, ModelField, ObjectSelections, Selections, ToManyField, ToOneField

__all__ = [
    "GraphQLASTWalker",
]


class GraphQLASTWalker:  # noqa: PLR0904
    """Class for walking the GraphQL AST and handling the different nodes."""

    def __init__(self, info: GQLInfo, model: type[Model]) -> None:
        self.info = info
        self.model = model

    def run(self) -> None:
        field_type = self.info.parent_type
        field_node = self.info.field_nodes[0]
        self.handle_query_class(field_type, field_node)

    def handle_query_class(self, parent_type: GraphQLObjectType, field_node: FieldNode) -> None:
        field_type: GraphQLCompositeType = self.get_field_type(parent_type, field_node)  # type: ignore[assignment]
        if field_node.selection_set is not None:
            self.handle_selections(field_type, field_node.selection_set.selections)

    def handle_selections(self, parent_type: GraphQLCompositeType, selections: Selections) -> None:
        if isinstance(parent_type, GraphQLInterfaceType | GraphQLUnionType):
            self.handle_abstract_type(parent_type, selections)
            return

        self.handle_object_type(parent_type, selections)  # type: ignore[arg-type]

    def handle_abstract_type(self, parent_type: GraphQLAbstractType, selections: Selections) -> None:
        if is_node_interface(parent_type):
            self.handle_node_interface(parent_type, selections)

        results = self.flatten_abstract_type_selections(selections)
        types_with_fragments: set[str] = set()

        for inline_fragment in results.inline_fragments:
            if should_skip_node(inline_fragment, self.info.variable_values):
                continue

            fragment_name = inline_fragment.type_condition.name.value
            types_with_fragments.add(fragment_name)

            fragment_type: GraphQLObjectType = self.info.schema.get_type(fragment_name)  # type: ignore[assignment]

            fragment_model = self.get_model(fragment_type)
            if fragment_model is None:  # pragma: no cover
                continue

            # Abstract types are only optimized in the top-level of the query.
            # In this case, we only optimize the fragment for the model that was set
            # in Optimizer initialization. Resolvers should run optimizer
            # for each returned concrete type separately.
            if fragment_model != self.model:
                continue

            fragment_selections = itertools.chain(inline_fragment.selection_set.selections, results.field_nodes)
            self.handle_selections(fragment_type, fragment_selections)

        # If there is no inline fragment for some concrete implementation of an interface,
        # but some fields have been selected from the interface, we still needs to fetch
        # the concrete implementations with the interface fields selected.
        if isinstance(parent_type, GraphQLInterfaceType) and results.field_nodes:
            undine_interface = get_undine_interface_type(parent_type)
            if undine_interface is None:
                return

            types_without_fragments = (
                query_type.__schema_name__
                for query_type in undine_interface.__concrete_implementations__()
                if query_type.__schema_name__ not in types_with_fragments
            )

            for fragment_name in types_without_fragments:
                fragment_type = self.info.schema.get_type(fragment_name)  # type: ignore[assignment]

                fragment_model = self.get_model(fragment_type)
                if fragment_model is None:  # pragma: no cover
                    continue

                if fragment_model != self.model:
                    continue

                self.handle_selections(fragment_type, results.field_nodes)

    def handle_node_interface(self, parent_type: GraphQLInterfaceType, selections: Selections) -> None: ...

    def handle_object_type(self, parent_type: GraphQLObjectType, selections: ObjectSelections) -> None:
        for selection in selections:
            if should_skip_node(selection, self.info.variable_values):
                continue

            if is_typename_metafield(selection):
                self.handle_typename_metafield(parent_type, selection)
                continue

            if isinstance(selection, FieldNode):
                with self.with_child_info(parent_type, selection):
                    self.handle_field_node(parent_type, selection)

            elif isinstance(selection, FragmentSpreadNode):
                fragment_definition = self.get_fragment_def(selection)
                fragment_selections: ObjectSelections = fragment_definition.selection_set.selections  # type: ignore[assignment]
                self.handle_object_type(parent_type, fragment_selections)

            else:  # pragma: no cover
                msg = f"Unhandled object selection node: '{selection}'"
                raise GraphQLOptimizerError(msg)

    def handle_field_node(self, parent_type: GraphQLObjectType, field_node: FieldNode) -> None:
        if is_connection(parent_type):
            self.handle_connection(parent_type, field_node)
            return

        if is_edge(parent_type):
            self.handle_edge_field(parent_type, field_node)
            return

        if is_page_info(parent_type):
            self.handle_page_info_field(parent_type, field_node)
            return

        field_model = self.get_model(parent_type)
        if field_model is None:  # pragma: no cover
            return

        with self.use_model(field_model):
            self.handle_model_field(parent_type, field_node)

    def handle_typename_metafield(self, parent_type: GraphQLCompositeType, field_node: FieldNode) -> None: ...

    def handle_connection(self, parent_type: GraphQLObjectType, field_node: FieldNode) -> None:
        field_type: GraphQLObjectType = self.get_field_type(parent_type, field_node)  # type: ignore[assignment]

        if field_node.name.value == undine_settings.TOTAL_COUNT_PARAM_NAME:
            self.handle_total_count(field_type, field_node)  # type: ignore[arg-type]
            return

        if field_node.selection_set is not None:
            self.handle_selections(field_type, field_node.selection_set.selections)

    def handle_edge_field(self, parent_type: GraphQLObjectType, field_node: FieldNode) -> None:
        field_type: GraphQLObjectType = self.get_field_type(parent_type, field_node)  # type: ignore[assignment]

        if field_node.selection_set is not None:
            self.handle_selections(field_type, field_node.selection_set.selections)

    def handle_page_info_field(self, parent_type: GraphQLObjectType, field_node: FieldNode) -> None: ...

    def handle_total_count(self, scalar: GraphQLScalarType, field_node: FieldNode) -> None: ...

    def handle_model_field(self, parent_type: GraphQLObjectType, field_node: FieldNode) -> None:
        field_name = to_snake_case(field_node.name.value)

        try:
            field: ModelField | None = get_model_field(model=self.model, lookup=field_name)
        except ModelFieldError:
            field = None

        if field is None:
            self.handle_custom_field(parent_type, field_node)
            return

        if not field.is_relation or is_relation_id(field, field_node):
            self.handle_normal_field(parent_type, field_node, field)  # type: ignore[arg-type]
            return

        if is_generic_foreign_key(field):
            self.handle_generic_foreign_key(parent_type, field_node, field)
            return

        if is_to_one(field):
            self.handle_to_one_field(parent_type, field_node, field)
            return

        if is_to_many(field):
            self.handle_to_many_field(parent_type, field_node, field)
            return

        msg = f"Unhandled field: '{field.name}'"  # pragma: no cover
        raise GraphQLOptimizerError(msg)  # pragma: no cover

    def handle_custom_field(
        self,
        parent_type: GraphQLObjectType,
        field_node: FieldNode,
    ) -> None: ...

    def handle_normal_field(
        self,
        parent_type: GraphQLObjectType,
        field_node: FieldNode,
        field: Field,
    ) -> None: ...

    def handle_to_one_field(
        self,
        parent_type: GraphQLObjectType,
        field_node: FieldNode,
        related_field: ToOneField,
    ) -> None:
        field_type: GraphQLObjectType = self.get_field_type(parent_type, field_node)  # type: ignore[assignment]
        if field_node.selection_set is not None:
            self.handle_selections(field_type, field_node.selection_set.selections)

    def handle_to_many_field(
        self,
        parent_type: GraphQLObjectType,
        field_node: FieldNode,
        related_field: ToManyField,
    ) -> None:
        field_type: GraphQLObjectType = self.get_field_type(parent_type, field_node)  # type: ignore[assignment]
        if field_node.selection_set is not None:
            self.handle_selections(field_type, field_node.selection_set.selections)

    def handle_generic_foreign_key(
        self,
        parent_type: GraphQLObjectType,
        field_node: FieldNode,
        related_field: GenericForeignKey,
    ) -> None:
        field_type: GraphQLUnionType = self.get_field_type(parent_type, field_node)  # type: ignore[assignment]
        if field_node.selection_set is not None:
            self.handle_selections(field_type, field_node.selection_set.selections)

    def get_model(self, object_type: GraphQLObjectType) -> type[Model] | None:
        query_type = get_undine_query_type(object_type)
        if query_type is None:
            return None
        return query_type.__model__

    def get_field_type(self, parent_type: GraphQLObjectType, field_node: FieldNode) -> GraphQLNamedOutputType:
        graphql_field = get_field_def(self.info.schema, parent_type, field_node)
        return get_underlying_type(graphql_field.type)  # type: ignore[return-value,type-var]

    def get_fragment_def(self, fragment_spread: FragmentSpreadNode) -> FragmentDefinitionNode:
        fragment_name = fragment_spread.name.value
        return self.info.fragments[fragment_name]

    def flatten_abstract_type_selections(self, selections: Selections) -> AbstractSelections:
        results = AbstractSelections()

        for selection in selections:
            if isinstance(selection, FieldNode):
                results.field_nodes.append(selection)

            elif isinstance(selection, InlineFragmentNode):
                results.inline_fragments.append(selection)

            elif isinstance(selection, FragmentSpreadNode):
                fragment_definition = self.get_fragment_def(selection)
                fragment_results = self.flatten_abstract_type_selections(fragment_definition.selection_set.selections)
                results.field_nodes += fragment_results.field_nodes
                results.inline_fragments += fragment_results.inline_fragments

        return results

    @contextmanager
    def use_model(self, model: type[Model]) -> Generator[None, Any, None]:
        orig_model = self.model
        try:
            self.model = model
            yield
        finally:
            self.model = orig_model

    @contextmanager
    def with_child_info(self, parent_type: GraphQLObjectType, field_node: FieldNode) -> Generator[None, None, None]:
        parent_info = self.info

        field_name = field_node.name.value
        field_def = get_field_def(self.info.schema, parent_type, field_node)
        key = get_field_entry_key(field_node)
        path = self.info.path.add_key(key, parent_type.name)

        info = GraphQLResolveInfo(
            field_name=field_name,
            field_nodes=[field_node],
            return_type=field_def.type,
            parent_type=parent_type,
            path=path,
            schema=self.info.schema,
            fragments=self.info.fragments,
            root_value=self.info.root_value,
            operation=self.info.operation,
            variable_values=self.info.variable_values,
            context=self.info.context,
            is_awaitable=self.info.is_awaitable,
        )

        try:
            self.info = info  # type: ignore[assignment]
            yield
        finally:
            self.info = parent_info



from __future__ import annotations

from contextlib import suppress
from types import FunctionType
from typing import Any, Never

from django.contrib.contenttypes.fields import GenericForeignKey, GenericRelation
from django.db.models import F, Q
from graphql import GraphQLFieldResolver, GraphQLID, GraphQLType, GraphQLWrappingType

from undine import Calculation, Field, InterfaceField, QueryType
from undine.converters import convert_to_field_resolver
from undine.dataclasses import LazyGenericForeignKey, LazyLambda, LazyRelation, TypeRef
from undine.exceptions import FunctionDispatcherError, RegistryMissingTypeError
from undine.pagination import OffsetPagination
from undine.relay import Connection
from undine.resolvers import (
    GlobalIDResolver,
    ModelAttributeResolver,
    ModelGenericForeignKeyResolver,
    ModelManyRelatedFieldResolver,
    ModelSingleRelatedFieldResolver,
    NestedConnectionResolver,
    NestedQueryTypeManyResolver,
    NestedQueryTypeSingleResolver,
)
from undine.resolvers.query import FieldFunctionResolver
from undine.typing import CombinableExpression, ModelField, ToManyField, ToOneField
from undine.utils.model_utils import get_model_field


@convert_to_field_resolver.register
def _(ref: FunctionType, **kwargs: Any) -> GraphQLFieldResolver:
    caller: Field = kwargs["caller"]
    return FieldFunctionResolver(func=ref, field=caller)


@convert_to_field_resolver.register
def _(_: ModelField, **kwargs: Any) -> GraphQLFieldResolver:
    caller: Field = kwargs["caller"]
    return ModelAttributeResolver(field=caller)


@convert_to_field_resolver.register
def _(_: ToOneField, **kwargs: Any) -> GraphQLFieldResolver:
    caller: Field = kwargs["caller"]
    return ModelSingleRelatedFieldResolver(field=caller)


@convert_to_field_resolver.register
def _(_: ToManyField, **kwargs: Any) -> GraphQLFieldResolver:
    caller: Field = kwargs["caller"]
    return ModelManyRelatedFieldResolver(field=caller)


@convert_to_field_resolver.register
def _(_: CombinableExpression | F | Q, **kwargs: Any) -> GraphQLFieldResolver:
    caller: Field = kwargs["caller"]
    return ModelAttributeResolver(field=caller)


@convert_to_field_resolver.register
def _(ref: LazyRelation, **kwargs: Any) -> GraphQLFieldResolver:
    try:
        value = ref.get_type()
    except RegistryMissingTypeError:
        value = ref.field

    return convert_to_field_resolver(value, **kwargs)


@convert_to_field_resolver.register
def _(ref: LazyGenericForeignKey, **kwargs: Any) -> GraphQLFieldResolver:
    return convert_to_field_resolver(ref.field, **kwargs)


@convert_to_field_resolver.register
def _(ref: LazyLambda, **kwargs: Any) -> GraphQLFieldResolver:
    return convert_to_field_resolver(ref.callback(), **kwargs)


@convert_to_field_resolver.register
def _(_: type[Calculation], **kwargs: Any) -> GraphQLFieldResolver:
    caller: Field = kwargs["caller"]
    return ModelAttributeResolver(field=caller, static=False)


@convert_to_field_resolver.register
def _(ref: TypeRef, **kwargs: Any) -> Never:
    caller: Field = kwargs["caller"]
    msg = f"Must define a custom resolve for '{caller.name}' since using python type '{ref.value}' as a reference."
    raise FunctionDispatcherError(msg)


@convert_to_field_resolver.register
def _(ref: GraphQLType, **kwargs: Any) -> Never:
    caller: Field = kwargs["caller"]
    msg = f"Must define a custom resolve for '{caller.name}' since using GraphQLType '{ref}' as a reference."
    raise FunctionDispatcherError(msg)


@convert_to_field_resolver.register
def _(ref: GraphQLWrappingType, **kwargs: Any) -> GraphQLFieldResolver:
    return convert_to_field_resolver(ref.of_type, **kwargs)


@convert_to_field_resolver.register
def _(_: GraphQLID, **kwargs: Any) -> GraphQLFieldResolver:
    caller: Field = kwargs["caller"]
    return GlobalIDResolver(typename=caller.query_type.__schema_name__)


@convert_to_field_resolver.register
def _(_: GenericForeignKey, **kwargs: Any) -> GraphQLFieldResolver:
    caller: Field = kwargs["caller"]
    return ModelGenericForeignKeyResolver(field=caller)


@convert_to_field_resolver.register
def _(_: GenericRelation, **kwargs: Any) -> GraphQLFieldResolver:
    caller: Field = kwargs["caller"]
    return ModelManyRelatedFieldResolver(field=caller)


@convert_to_field_resolver.register
def _(ref: type[QueryType], **kwargs: Any) -> GraphQLFieldResolver:
    caller: Field = kwargs["caller"]
    if caller.many:
        return NestedQueryTypeManyResolver(field=caller, query_type=ref)
    return NestedQueryTypeSingleResolver(field=caller, query_type=ref)


@convert_to_field_resolver.register
def _(ref: Connection, **kwargs: Any) -> GraphQLFieldResolver:
    caller: Field = kwargs["caller"]
    return NestedConnectionResolver(connection=ref, field=caller)


@convert_to_field_resolver.register
def _(ref: OffsetPagination, **kwargs: Any) -> GraphQLFieldResolver:
    caller: Field = kwargs["caller"]
    return NestedQueryTypeManyResolver(query_type=ref.query_type, field=caller)


@convert_to_field_resolver.register
def _(ref: InterfaceField, **kwargs: Any) -> GraphQLFieldResolver:
    # Attempt to find a resolver for the GraphQL scalar (e.g. ID).
    with suppress(FunctionDispatcherError):
        return convert_to_field_resolver(ref.output_type, **kwargs)

    caller: Field = kwargs["caller"]
    field = get_model_field(model=caller.query_type.__model__, lookup=caller.field_name)
    return convert_to_field_resolver(field, **kwargs)



from __future__ import annotations

import json
from inspect import cleandoc

from undine.exceptions import GraphQLScalarInvalidValueError

from ._definition import ScalarType

__all__ = [
    "GraphQLJSON",
    "json_scalar",
]


json_scalar: ScalarType[dict, dict] = ScalarType(
    name="JSON",
    description=cleandoc(
        """
        Represents a JSON serializable object.
        Maps to the Python `dict` type.
        """
    ),
    specified_by_url="https://datatracker.ietf.org/doc/html/rfc8259",
)

GraphQLJSON = json_scalar.as_graphql_scalar()


@json_scalar.serialize.register
@json_scalar.parse.register
def _(value: dict) -> dict:
    return value


@json_scalar.serialize.register
@json_scalar.parse.register
def _(value: str) -> dict:
    value = json.loads(value)
    if not isinstance(value, dict):
        raise GraphQLScalarInvalidValueError(typename="JSON")

    return value


@json_scalar.serialize.register
def _(value: bytes) -> dict:
    value = json.loads(value)
    if not isinstance(value, dict):
        raise GraphQLScalarInvalidValueError(typename="JSON")

    return value



from __future__ import annotations

import inspect
from typing import TYPE_CHECKING, Any

from django.db.models import QuerySet
from graphql import GraphQLResolveInfo, Undefined

from undine.dataclasses import Parameter
from undine.exceptions import MissingFunctionAnnotationsError, MissingFunctionReturnTypeError, NoFunctionParametersError
from undine.settings import undine_settings
from undine.typing import GQLInfo
from undine.utils.reflection import get_origin_or_noop, get_signature, is_subclass

if TYPE_CHECKING:
    from collections.abc import Callable
    from types import FunctionType


__all__ = [
    "parse_first_param_type",
    "parse_parameters",
    "parse_return_annotation",
]


def parse_parameters(func: FunctionType | Callable[..., Any], *, depth: int = 0) -> list[Parameter]:
    """
    Parse function arguments, type hints, and default values into parameters.
    Only parses arguments that can be converted to GraphQL arguments.

    :param func: Function to parse.
    :param depth: How many function calls deep is the code calling this method compared to the parsed function?
                  This can be omitted if the function's signature has been parsed previously.
    """
    sig = get_signature(func, depth=depth + 1)

    missing: list[str] = []
    parameters: list[Parameter] = []

    for i, param in enumerate(sig.parameters.values()):
        # 'self' and 'cls' parameters are special and thus skipped.
        if i == 0 and param.name in {"self", "cls", undine_settings.RESOLVER_ROOT_PARAM_NAME}:
            continue

        # Don't include '*args' and '**kwargs' parameters, as they are not supported by GraphQL.
        if param.kind in {inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD}:
            continue

        if param.annotation is inspect.Parameter.empty:
            missing.append(param.name)
            continue

        if get_origin_or_noop(param.annotation) in {GraphQLResolveInfo, GQLInfo}:
            continue

        if is_subclass(param.annotation, QuerySet):
            continue

        parameters.append(
            Parameter(
                name=param.name,
                annotation=param.annotation,
                default_value=param.default if param.default is not inspect.Parameter.empty else Undefined,
            ),
        )

    if missing:
        raise MissingFunctionAnnotationsError(missing=missing, func=func) from None

    return parameters


def parse_first_param_type(func: FunctionType | Callable[..., Any], *, depth: int = 0) -> type:
    """
    Get the type of the first parameter of the given function.

    :param func: Function to parse.
    :param depth: How many function calls deep is the code calling this method compared to the parsed function?
                  This can be omitted if the function's signature has been parsed previously.
    """
    parameters = parse_parameters(func, depth=depth + 1)

    param_type = next((param.annotation for param in parameters), Undefined)
    if param_type is Undefined:
        raise NoFunctionParametersError(func=func)

    return param_type


def parse_return_annotation(func: FunctionType | Callable[..., Any], *, depth: int = 0) -> type:
    """
    Parse the return annotation of the given function.

    :param func: Function to parse.
    :param depth: How many function calls deep is the code calling this method compared to the parsed function?
                  This can be omitted if the function's signature has been parsed previously.
    """
    sig = get_signature(func, depth=depth + 1)

    if sig.return_annotation is inspect.Parameter.empty:
        raise MissingFunctionReturnTypeError(func=func) from None

    return sig.return_annotation



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from django.core.exceptions import ValidationError
from django.db.models import CharField

from undine.utils.text import comma_sep_str

if TYPE_CHECKING:
    from collections.abc import Sequence

    from django.db.models import TextChoices


__all__ = [
    "TextChoicesField",
]


class TextChoicesField(CharField):
    """Choice field that offers more consistent naming for enums in the GraphQL Schema."""

    default_error_messages = {
        "invalid": "`%(value)s` is not a member of the `%(enum_name)s` enum. Choices are: %(choices)s.",
    }

    def __init__(self, choices_enum: type[TextChoices], **kwargs: Any) -> None:
        """
        A field for enums that use TextChoices.
        Automatically adds 'max_length' validator based on the enum values.
        Values fetched from the database are converted to the enum type.

        :param choices_enum: `TextChoices` class to use for the field.
        :param kwargs: Keyword arguments to pass to the `CharField` constructor.
        """
        self.choices_enum = choices_enum
        kwargs["choices"] = choices_enum.choices
        kwargs["max_length"] = max(len(val) for val in choices_enum.values)
        super().__init__(**kwargs)

    def deconstruct(self) -> tuple[str, str, Sequence[str], dict[str, Any]]:
        """Returns a tuple with enough information to recreate the field."""
        name, path, args, kwargs = super().deconstruct()
        kwargs["choices_enum"] = self.choices_enum
        return name, path, args, kwargs

    def to_python(self, value: Any) -> Any:
        """Converts the given value into the correct Python object for this field."""
        # The database can return `None` even if this field is not nullable
        # in case the field selected through a relation where the relationship itself is null.
        #
        # When saving to database, we can return `None` here since the database
        # will throw an `IntegrityError` if the field is not nullable.
        if value is None:
            return None

        try:
            return self.choices_enum(value)
        except ValueError as error:
            params = {
                "value": value,
                "enum_name": self.choices_enum.__name__,
                "choices": comma_sep_str(self.choices_enum.values, last_sep="and", quote=True),
            }
            error_dict = {self.attname: self.error_messages["invalid"] % params}
            raise ValidationError(error_dict, code="invalid") from error

    def from_db_value(self, value: Any, expression: Any, connection: Any) -> Any:
        """Converts a value as returned by the database to a Python object."""
        return self.to_python(value)

    def get_prep_value(self, value: Any) -> Any:
        """Converts the given value into a value suitable for the database."""
        return self.to_python(value)



from __future__ import annotations

from types import FunctionType
from typing import Any

from django.contrib.contenttypes.fields import GenericForeignKey
from django.db.models import Model
from graphql import Undefined

from undine import Input, MutationType
from undine.converters import is_input_required
from undine.dataclasses import LazyLambda, TypeRef
from undine.exceptions import ModelFieldError
from undine.parsers import parse_is_nullable, parse_parameters
from undine.typing import ModelField, MutationKind
from undine.utils.graphql.utils import is_non_null_default_value
from undine.utils.model_utils import get_model_field, has_default, is_to_many


@is_input_required.register
def _(ref: ModelField, **kwargs: Any) -> bool:
    caller: Input = kwargs["caller"]

    is_primary_key = bool(getattr(ref, "primary_key", False))
    is_nullable = bool(getattr(ref, "null", True))
    is_to_many_field = is_to_many(ref)
    has_non_null_default_value = is_non_null_default_value(caller.default_value)
    has_field_default = has_default(ref)

    match caller.mutation_type.__kind__:
        case MutationKind.create:
            if is_to_many_field:
                return False
            if is_nullable:
                return False
            if has_non_null_default_value:
                return True
            return not has_field_default

        case MutationKind.update | MutationKind.delete:
            return is_primary_key

        case MutationKind.custom:
            return True

        case MutationKind.related:
            return False

        case _:  # pragma: no cover
            return False


@is_input_required.register
def _(_: type[Model], **kwargs: Any) -> bool:
    caller: Input = kwargs["caller"]

    try:
        field = get_model_field(model=caller.mutation_type.__model__, lookup=caller.field_name)
    except ModelFieldError:
        return True

    return is_input_required(field, **kwargs)


@is_input_required.register
def _(ref: TypeRef, **kwargs: Any) -> bool:
    caller: Input = kwargs["caller"]

    # GraphQL doesn't differentiate between null and required.
    nullable = parse_is_nullable(ref.value, is_input=True, total=ref.total)
    if nullable:
        return False

    if is_non_null_default_value(caller.default_value):
        return True

    return caller.mutation_type.__kind__ in {MutationKind.create, MutationKind.custom}


@is_input_required.register
def _(_: LazyLambda, **kwargs: Any) -> bool:
    return False


@is_input_required.register
def _(ref: FunctionType, **kwargs: Any) -> bool:
    parameters = parse_parameters(ref)
    first_param_default_value = next((param.default_value for param in parameters), Undefined)
    return is_non_null_default_value(first_param_default_value)


@is_input_required.register
def _(_: type[MutationType], **kwargs: Any) -> bool:
    caller: Input = kwargs["caller"]
    field = get_model_field(model=caller.mutation_type.__model__, lookup=caller.field_name)
    return is_input_required(field, caller=caller)


@is_input_required.register
def _(_: GenericForeignKey, **kwargs: Any) -> bool:
    caller: Input = kwargs["caller"]
    return caller.mutation_type.__kind__ == MutationKind.create



from __future__ import annotations

from http import HTTPStatus
from typing import TYPE_CHECKING

from graphql import ExecutionResult, GraphQLError

from undine.exceptions import GraphQLErrorGroup, GraphQLRequestDecodingError
from undine.http.utils import graphql_result_response, parse_json_body, require_persisted_documents_request
from undine.settings import undine_settings
from undine.utils.graphql.utils import get_error_execution_result

from .utils import parse_document_map, register_persisted_documents

if TYPE_CHECKING:
    from undine.typing import DjangoRequestProtocol, DjangoResponseProtocol

__all__ = [
    "persisted_documents_view",
]


@require_persisted_documents_request
def persisted_documents_view(request: DjangoRequestProtocol) -> DjangoResponseProtocol:
    """
    View for registering persisted documents.
    Users should add permission checks.
    """
    try:
        json_data = parse_json_body(request.body)

    except GraphQLRequestDecodingError as error:
        result = get_error_execution_result(error)
        return graphql_result_response(result, status=HTTPStatus.BAD_REQUEST)

    try:
        document_map = parse_document_map(json_data)

    except (GraphQLError, GraphQLErrorGroup) as error:
        result = get_error_execution_result(error)
        return graphql_result_response(result, status=HTTPStatus.BAD_REQUEST)

    if undine_settings.PERSISTED_DOCUMENTS_PERMISSION_CALLBACK is not None:
        try:
            undine_settings.PERSISTED_DOCUMENTS_PERMISSION_CALLBACK(request, document_map)

        except (GraphQLError, GraphQLErrorGroup) as error:
            result = get_error_execution_result(error)
            return graphql_result_response(result, status=HTTPStatus.FORBIDDEN)

    try:
        document_id_map = register_persisted_documents(document_map)

    except (GraphQLError, GraphQLErrorGroup) as error:
        result = get_error_execution_result(error)
        return graphql_result_response(result, status=HTTPStatus.BAD_REQUEST)

    result = ExecutionResult(data={"documents": document_id_map})
    return graphql_result_response(result)



from __future__ import annotations

import inspect
from collections.abc import Hashable
from types import FunctionType, UnionType
from typing import TYPE_CHECKING, Any, Generic, Literal, Union, get_origin

from graphql import Undefined

from undine.dataclasses import DispatchImplementations
from undine.exceptions import (
    FunctionDispatcherImplementationNotFoundError,
    FunctionDispatcherImproperLiteralError,
    FunctionDispatcherNoArgumentAnnotationError,
    FunctionDispatcherNoArgumentsError,
    FunctionDispatcherRegistrationError,
    FunctionDispatcherUnknownArgumentError,
)
from undine.typing import Lambda, LiteralArg, T

from .reflection import (
    can_be_literal_arg,
    get_flattened_generic_params,
    get_instance_name,
    get_non_null_type,
    get_origin_or_noop,
    get_signature,
    is_lambda,
)

if TYPE_CHECKING:
    from collections.abc import Callable, Generator

    from undine.typing import DispatchProtocol

__all__ = [
    "FunctionDispatcher",
]


class FunctionDispatcher(Generic[T]):
    """
    A class that holds different implementations for a function
    based on the function's first argument. Different implementations can be added with the `register` method.
    Use implementations by calling the instance with a single positional argument and any number of keyword arguments.
    If no implementation is found, an error will be raised.
    """

    def __init__(self, *, wrapper: Callable[[DispatchProtocol[T]], DispatchProtocol[T]] | None = None) -> None:
        """
        Create a new FunctionDispatcher. Must be added to a variable before use!

        :param wrapper: A function that wraps all implemented functions for performing additional logic.
        """
        self.__name__ = get_instance_name()
        self.implementations: DispatchImplementations[T] = DispatchImplementations()
        self.wrapper = wrapper
        self.default: DispatchProtocol[T] = Undefined  # type: ignore[assignment]

    def __class_getitem__(cls, key: T) -> FunctionDispatcher[T]:
        """Adds typing information when used like this: `foo = FunctionDispatcher[T]()`."""
        return cls  # type: ignore[return-value]

    def __call__(self, value: Any, /, **kwargs: Any) -> T:
        """Find the implementation for the given key and call it with the given keyword arguments."""
        key = get_non_null_type(value)
        implementation = self[value]
        return implementation(key, **kwargs)

    def __getitem__(self, original_key: Any) -> DispatchProtocol[T]:
        """Find the implementation for the given key."""
        non_null_key = get_non_null_type(original_key)
        key = get_origin_or_noop(non_null_key)

        if is_lambda(key):
            impl = self.implementations.types.get(Lambda)
            if impl is not None:
                return impl

        elif can_be_literal_arg(key):
            impl = self.implementations.literals.get(key)
            if impl is not None:
                return impl

        if isinstance(key, type):
            section: dict[Any, DispatchProtocol] = self.implementations.types
            cls: type = get_origin_or_noop(key)

        else:
            section = self.implementations.instances
            cls = type(key)

            if isinstance(key, Hashable):
                impl = section.get(key)
                if impl is not None:
                    return impl

        for mro_cls in cls.__mro__:
            impl = section.get(mro_cls)
            if impl is not None:
                return impl

        if self.default is not Undefined:
            return self.default

        raise FunctionDispatcherImplementationNotFoundError(name=self.__name__, key=key, cls=cls)

    def __contains__(self, value: Any) -> bool:
        try:
            self[value]
        except FunctionDispatcherImplementationNotFoundError:
            return False
        return True

    def register(self, func: Callable[..., T]) -> Callable[..., T]:
        """Register the given function as an implementation for its first argument's type."""
        if not isinstance(func, FunctionType):
            raise FunctionDispatcherRegistrationError(name=self.__name__, value=func)

        annotation = self._first_param_type(func, depth=1)

        if annotation is Any:
            self.default = self.wrapper(func) if self.wrapper else func
            return func

        if annotation in {Lambda, type}:
            self.implementations.types[annotation] = self.wrapper(func) if self.wrapper else func
            return func

        origin = get_origin(annotation)

        # Example: "str" or "int"
        if not origin:
            self.implementations.instances[annotation] = self.wrapper(func) if self.wrapper else func
            return func

        for section, arg in self._iter_args(annotation):
            implementations = getattr(self.implementations, section)
            implementations[arg] = self.wrapper(func) if self.wrapper else func

        return func

    def _first_param_type(self, func: FunctionType, *, depth: int = 0) -> Any:
        """Get the type of the first parameter of the given function."""
        sig = get_signature(func, depth=depth + 1)

        try:
            annotation = next(param.annotation for param in sig.parameters.values())
        except StopIteration as error:
            raise FunctionDispatcherNoArgumentsError(func_name=func.__name__, name=self.__name__) from error

        if annotation is inspect.Parameter.empty:
            raise FunctionDispatcherNoArgumentAnnotationError(func_name=func.__name__, name=self.__name__)

        return annotation

    def _iter_args(self, annotation: Any) -> Generator[tuple[str, Any], None, None]:
        origin = get_origin(annotation)

        for arg in get_flattened_generic_params(annotation):
            arg_origin = get_origin(arg)

            # Example: "str | int" or "Union[str, int]"
            if origin in {UnionType, Union}:
                if arg_origin is not None:
                    yield from self._iter_args(arg)
                else:
                    yield "instances", arg

            # Example: "type[str]" or "type[str | int]"
            elif origin is type:
                if arg_origin is not None:
                    yield from (("types", ann) for _, ann in self._iter_args(arg))
                else:
                    yield "types", arg

            # Example: Literal["foo", "bar"]
            elif origin is Literal:
                if not isinstance(arg, LiteralArg):
                    raise FunctionDispatcherImproperLiteralError(arg=arg)
                yield "literals", arg

            else:
                raise FunctionDispatcherUnknownArgumentError(annotation=annotation)



from __future__ import annotations

import asyncio
import uuid
from abc import ABC, abstractmethod
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Generic

from django.db.models.signals import post_save, pre_delete

from undine.exceptions import GraphQLSubscriptionTimeoutError
from undine.typing import T, TModel

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator

    from django.dispatch import Signal

    from undine import QueryType
    from undine.typing import PostDeleteParams, PostSaveParams

__all__ = [
    "ModelCreateSubscription",
    "ModelDeleteSubscription",
    "ModelSaveSubscription",
    "ModelUpdateSubscription",
    "SignalSubscriber",
    "SignalSubscription",
]


class SignalSubscription(ABC, Generic[T]):
    """A subscription that forwards data from a signal."""

    def __init__(
        self,
        sender: Any,
        *,
        dispatch_uid: str | None = None,
        description: str | None = None,
        timeout: float | None = None,
    ) -> None:
        """
        Create a new subscription.

        :param sender: The model class to subscribe to.
        :param dispatch_uid: The dispatch uid for the signal.
        :param description: The description for the subscription.
        :param timeout: How long to wait between signals before timing out.
        """
        self.sender = sender
        self.dispatch_uid = dispatch_uid
        self.description = description
        self.timeout = timeout

        self.subscribers: dict[uuid.UUID, SignalSubscriber] = {}

        self.signal.connect(self.receiver, sender=sender, dispatch_uid=dispatch_uid)

    @property
    @abstractmethod
    def signal(self) -> Signal:
        """The signal to subscribe to."""

    @abstractmethod
    def transform(self, params: dict[str, Any]) -> T:
        """Transform the given event data into the desired output."""

    def filter(self, params: dict[str, Any]) -> bool:
        """Should the given event be filtered out?"""
        return False

    def process(self, params: dict[str, Any]) -> dict[str, Any]:
        """Process the given signal data before handing it out to subscribers."""
        return params

    def create_subscriber(self) -> SignalSubscriber[T]:
        return SignalSubscriber(self)

    def receiver(self, *args: Any, **kwargs: Any) -> None:
        """Receiver for the Django signal."""
        # Some signals might send the 'sender' argument as a positional argument
        if args:
            kwargs["sender"] = args[0]

        data = self.process(kwargs)
        for subscriber in self.subscribers.values():
            subscriber.events.put_nowait(data)


class SignalSubscriber(Generic[T]):
    """Subscriber that receives events from a signal subscription."""

    def __init__(self, subscription: SignalSubscription) -> None:
        """
        Create a new subscriber.

        :param subscription: The subscription this subscriber is for.
        """
        self.subscription = subscription
        self.events: asyncio.Queue[dict[str, Any]] = asyncio.Queue()

    async def subscribe(self) -> AsyncGenerator[T, None]:
        """Begin receiving events from the subscription."""
        key = uuid.uuid4()
        self.subscription.subscribers[key] = self
        try:
            while True:
                try:
                    event = await asyncio.wait_for(self.events.get(), timeout=self.subscription.timeout)
                except TimeoutError as error:
                    raise GraphQLSubscriptionTimeoutError from error

                if self.subscription.filter(event):
                    continue

                yield self.subscription.transform(event)
        finally:
            self.subscription.subscribers.pop(key, None)


class QueryTypeSignalSubscription(SignalSubscription[TModel], ABC):
    """Signal subscription for returning model instances through a QueryType."""

    def __init__(
        self,
        query_type: type[QueryType],
        /,
        *,
        dispatch_uid: str | None = None,
        description: str | None = None,
        timeout: float | None = None,
    ) -> None:
        """
        Create a new signal subscription that resolves using a QueryType.

        :param query_type: The QueryType to use for the subscription.
        :param dispatch_uid: The dispatch uid for the signal.
        :param description: The description for the subscription.
        :param timeout: How long to wait between signals before timing out.
        """
        self.query_type = query_type
        super().__init__(
            sender=query_type.__model__,
            dispatch_uid=dispatch_uid,
            description=description,
            timeout=timeout,
        )


class ModelSaveSubscription(QueryTypeSignalSubscription[TModel]):
    """Subscription that sends an event after a model instance has been saved."""

    signal = post_save

    def transform(self, params: PostSaveParams[TModel]) -> TModel:
        return params["instance"]


class ModelCreateSubscription(ModelSaveSubscription[TModel]):
    """Subscription that sends an event after a model instance has been created."""

    def filter(self, params: PostSaveParams[TModel]) -> bool:
        return not params["created"]


class ModelUpdateSubscription(ModelSaveSubscription[TModel]):
    """Subscription that sends an event after a model instance has been updated."""

    def filter(self, params: PostSaveParams[TModel]) -> bool:
        return params["created"]


class ModelDeleteSubscription(QueryTypeSignalSubscription[TModel]):
    """Subscription that sends an event before a model instance has been deleted."""

    signal = pre_delete

    def process(self, params: PostDeleteParams[TModel]) -> PostDeleteParams[TModel]:
        # It's possible that the instance is no longer in the database when
        # a subscriber receives the event. Therefore, make a deepcopy of it
        # so that its pk is still available when querying for the output.
        # However, relations cannot be queried since they are not prefetched.
        params["instance"] = deepcopy(params["instance"])
        return params

    def transform(self, params: PostDeleteParams[TModel]) -> TModel:
        return params["instance"]



from __future__ import annotations

import asyncio
import dataclasses
import json
import uuid
from contextlib import suppress
from typing import TYPE_CHECKING, Any, Self

from asgiref.typing import WebSocketConnectEvent, WebSocketReceiveEvent
from channels.auth import AuthMiddlewareStack

from undine.exceptions import WebSocketConnectionClosedError
from undine.integrations.channels import GraphQLWebSocketConsumer
from undine.settings import undine_settings
from undine.typing import CompleteMessage, ConnectionInitMessage, SubscribeMessage

if TYPE_CHECKING:
    from asgiref.typing import ASGISendEvent, WebSocketCloseEvent, WebSocketDisconnectEvent

    from undine.typing import (
        ClientMessage,
        ConnectionAckMessage,
        ErrorMessage,
        NextMessage,
        ServerMessage,
        WebSocketASGIScope,
    )


__all__ = [
    "TestWebSocket",
]


@dataclasses.dataclass(slots=True, kw_only=True)
class TestWebSocket:
    """A testing utility for sending and receiving messages from the given consumer."""

    scope: WebSocketASGIScope
    """ASGI scope for the WebSocket request."""

    consumer: GraphQLWebSocketConsumer = dataclasses.field(default_factory=GraphQLWebSocketConsumer)
    """Consumer for the WebSocket."""

    task: asyncio.Task | None = None
    """Consumer task with AuthMiddlewareStack applied. Added on context enter."""

    messages: asyncio.Queue[ClientMessage] = dataclasses.field(default_factory=asyncio.Queue)
    """Messages placed here are picked up by the consumer."""

    responses: asyncio.Queue[ServerMessage | None] = dataclasses.field(default_factory=asyncio.Queue)
    """Messages from the consumer are placed here."""

    accepted: asyncio.Event = dataclasses.field(default_factory=asyncio.Event)
    """Set if the connection is accepted after 'ConnectionInit' message is sent."""

    close_event: WebSocketCloseEvent | None = None
    """Set if the connection is closed."""

    async def connection_init(
        self,
        payload: dict[str, Any] | None = None,
        *,
        timeout: float | None = 3,
    ) -> ConnectionAckMessage:
        """
        Send a ConnectionInit message to the server.
        This is required before sending any other messages.

        :param payload: The connection init payload.
        :param timeout: Timeout in seconds for receiving the response.
        """
        message = ConnectionInitMessage(type="connection_init")
        if payload is not None:
            message["payload"] = payload
        return await self.send_and_receive(message=message, timeout=timeout)  # type: ignore[return-value]

    async def subscribe(
        self,
        payload: dict[str, Any],
        *,
        operation_id: str | None = None,
        timeout: float | None = 3,
    ) -> NextMessage | ErrorMessage:
        """
        Create a subscription for the given payload.

        If Next message is received, should wait for subsequent messages until receiving a Complete message.
        If Error message is received (at any point), no more messages will be sent.

        :param payload: The subscription payload.
        :param operation_id: The ID of the subscription operation.
        :param timeout: Timeout in seconds for receiving the first response.
        """
        operation_id = operation_id or str(uuid.uuid4())
        message = SubscribeMessage(type="subscribe", id=operation_id, payload=payload)
        return await self.send_and_receive(message=message, timeout=timeout)  # type: ignore[return-value]

    async def unsubscribe(self, *, operation_id: str) -> None:
        """
        Unsubscribes from a subscription early.

        Requires that some part of the subscription delays execution with
        some form of `loop.call_later`, (e.g. `asyncio.sleep` with a positive delay).
        Does not guarantee how many subscription messages will still be sent.
        Should manually fetch the operation from `consumer.handler.operations` and
        await its `.task` to complete with a `asyncio.CancelledError`.

        :param operation_id: The ID of the subscription operation.
        """
        message = CompleteMessage(type="complete", id=operation_id)
        await self.send(message=message)

    async def send_and_receive(self, message: ClientMessage, *, timeout: float | None = 3) -> ServerMessage:
        """
        Send a message to the server and wait for the response.

        :param message: The message to send.
        :param timeout: Timeout in seconds for receiving the response.
        :raises WebSocketConnectionClosedError: Connection is closed.
        :raises TimeoutError: Timeout reached.
        """
        await self.send(message=message)
        return await self.receive(timeout=timeout)

    async def send(self, message: ClientMessage) -> None:
        """
        Send a message to the server.

        :param message: The message to send.
        :raises WebSocketConnectionClosedError: Connection is closed.
        """
        if self.close_event is not None:
            raise WebSocketConnectionClosedError(reason=self.close_event["reason"], code=self.close_event["code"])

        await self.messages.put(message)

    async def receive(self, *, timeout: float | None = 3) -> ServerMessage:
        """
        Receive a message from the server.

        :param timeout: Timeout for receiving the message.
        :raises WebSocketConnectionClosedError: Connection is closed.
        :raises TimeoutError: Timeout reached.
        """
        if self.close_event is not None:
            raise WebSocketConnectionClosedError(reason=self.close_event["reason"], code=self.close_event["code"])

        timeout = None if undine_settings.TESTING_CLIENT_NO_ASYNC_TIMEOUT else timeout

        try:
            data = await asyncio.wait_for(self.responses.get(), timeout)
        except TimeoutError as error:
            msg = "Timeout waiting for message from server."
            raise TimeoutError(msg) from error

        if self.close_event is not None:
            raise WebSocketConnectionClosedError(reason=self.close_event["reason"], code=self.close_event["code"])

        if data is None:
            raise WebSocketConnectionClosedError

        return data

    async def __aenter__(self) -> Self:
        """Start the WebSocket connection."""
        self.accepted.clear()
        self.close_event = None

        stack = AuthMiddlewareStack(self.consumer)
        self.task = asyncio.create_task(stack(self.scope, self._to_consumer, self._from_consumer))

        await self.accepted.wait()
        return self

    async def __aexit__(self, *args: object) -> None:
        """Terminate the WebSocket connection."""
        # Cancel pending tasks in the consumer
        await self.consumer.handler.disconnect()

        # Shouldn't happen, but just in case.
        if self.task is None:  # pragma: no cover
            return

        # Make sure the task is done.
        if not self.task.done():
            self.task.cancel()

        timeout = None if undine_settings.TESTING_CLIENT_NO_ASYNC_TIMEOUT else 3

        with suppress(BaseException):
            # IDK why, but we need to wrap the task with `asyncio.wait` or we might wait here forever...
            await asyncio.wait(self.task, timeout=timeout)

    async def _to_consumer(self) -> WebSocketConnectEvent | WebSocketReceiveEvent | WebSocketDisconnectEvent:
        """
        Send an event to the consumer. Called automatically by `self.task`.
        Will wait for new messages until the connection is closed, or specified timeout is reached.

        :returns: The event to send to the consumer.
        :raises WebSocketConnectionClosedError: Connection is closed.
        :raises TimeoutError: Timeout reached.
        """
        if not self.accepted.is_set():
            return WebSocketConnectEvent(type="websocket.connect")

        if self.close_event is not None:
            raise WebSocketConnectionClosedError(reason=self.close_event["reason"], code=self.close_event["code"])

        timeout = None if undine_settings.TESTING_CLIENT_NO_ASYNC_TIMEOUT else 3

        try:
            message = await asyncio.wait_for(self.messages.get(), timeout=timeout)
        except TimeoutError as error:
            msg = "Timeout waiting for message from client."
            raise TimeoutError(msg) from error

        return WebSocketReceiveEvent(type="websocket.receive", text=json.dumps(message), bytes=None)

    async def _from_consumer(self, event: ASGISendEvent) -> None:
        """
        Event received from the consumer. Called automatically by `self.task`.

        :param event: The event received from the consumer.
        :raises RuntimeError: Unexpected event.
        """
        match event["type"]:
            case "websocket.accept":
                self.accepted.set()

            case "websocket.close":
                # Set 'None' to prevent receive from timing out.
                await self.responses.put(None)
                self.close_event = event

            case "websocket.send" if event["text"] is not None:
                await self.responses.put(json.loads(event["text"]))

            case _:
                # Set 'None' to prevent receive from timing out.
                await self.responses.put(None)
                msg = f"Unexpected event: {json.dumps(event)}"
                raise RuntimeError(msg)



from __future__ import annotations

from typing import TYPE_CHECKING

from graphql import GraphQLError

from undine.exceptions import GraphQLErrorGroup
from undine.execution import execute_graphql_http_async, execute_graphql_http_sync
from undine.http.utils import graphql_result_response, require_graphql_request
from undine.parsers import GraphQLRequestParamsParser
from undine.utils.graphql.utils import get_error_execution_result

if TYPE_CHECKING:
    from undine.typing import DjangoRequestProtocol, DjangoResponseProtocol

__all__ = [
    "graphql_view_async",
    "graphql_view_sync",
]


@require_graphql_request
def graphql_view_sync(request: DjangoRequestProtocol) -> DjangoResponseProtocol:
    """A sync view for GraphQL requests."""
    try:
        params = GraphQLRequestParamsParser.run(request)
    except (GraphQLError, GraphQLErrorGroup) as error:
        result = get_error_execution_result(error)
    else:
        result = execute_graphql_http_sync(params, request)

    return graphql_result_response(result, content_type=request.response_content_type)


@require_graphql_request
async def graphql_view_async(request: DjangoRequestProtocol) -> DjangoResponseProtocol:
    """A async view for GraphQL requests."""
    try:
        params = GraphQLRequestParamsParser.run(request)
    except (GraphQLError, GraphQLErrorGroup) as error:
        result = get_error_execution_result(error)
    else:
        result = await execute_graphql_http_async(params, request)

    return graphql_result_response(result, content_type=request.response_content_type)



from __future__ import annotations

from django.core.validators import validate_ipv4_address

from ._definition import ScalarType

__all__ = [
    "GraphQLIPv4",
    "ipv4_scalar",
]


ipv4_scalar: ScalarType[str, str] = ScalarType(
    name="IPv4",
    description="Represents a valid IPv4 address.",
    specified_by_url="https://datatracker.ietf.org/doc/html/rfc791",
)

GraphQLIPv4 = ipv4_scalar.as_graphql_scalar()


@ipv4_scalar.serialize.register
@ipv4_scalar.parse.register
def _(value: str) -> str:
    validate_ipv4_address(value)
    return value



from __future__ import annotations

from functools import cache
from typing import TYPE_CHECKING, Any

from graphql import (
    GraphQLError,
    GraphQLInputObjectType,
    NonNullTypeNode,
    NullValueNode,
    ValidationRule,
    VariableNode,
    get_named_type,
)

if TYPE_CHECKING:
    from graphql import GraphQLNamedType, ObjectValueNode, ValidationContext, VariableDefinitionNode, VisitorAction


__all__ = [
    "OneOfInputObjectTypeRule",
    "core_implements_one_of_directive",
    "get_one_of_input_object_type_extension",
    "is_one_of_input_object",
    "validate_one_of_input_object_variable_value",
]


class OneOfInputObjectTypeRule(ValidationRule):
    """
    Forwards `@oneOf` directive validation that exists in `ValuesOfCorrectTypeRule` from GraphQL core `v3.2.7`

    See: graphql/validation/rules/values_of_correct_type.py
    """

    def __init__(self, context: ValidationContext) -> None:
        super().__init__(context)
        self.variable_definitions: dict[str, VariableDefinitionNode] = {}

    def enter_operation_definition(self, *_args: Any) -> None:
        self.variable_definitions.clear()

    def enter_variable_definition(self, definition: VariableDefinitionNode, *_args: Any) -> None:
        self.variable_definitions[definition.variable.name.value] = definition

    def enter_object_value(self, node: ObjectValueNode, *_args: Any) -> VisitorAction:
        named_type: GraphQLNamedType | None = get_named_type(self.context.get_input_type())
        if named_type is None or not isinstance(named_type, GraphQLInputObjectType):
            return None

        if not is_one_of_input_object(named_type):
            return None

        typename = named_type.name

        if len(node.fields) != 1:
            msg = f"OneOf Input Object '{typename}' must specify exactly one key."
            self.context.report_error(GraphQLError(msg, node))
            return None

        key = node.fields[0].name.value
        value = node.fields[0].value

        if isinstance(value, NullValueNode):
            msg = f"Field '{typename}.{key}' must be non-null."
            self.context.report_error(GraphQLError(msg, node))
            return None

        if isinstance(value, VariableNode):
            variable_name = value.name.value
            definition = self.variable_definitions[variable_name]

            if not isinstance(definition.type, NonNullTypeNode):
                msg = f"Variable '{variable_name}' must be non-nullable to be used for OneOf Input Object '{typename}'."
                self.context.report_error(GraphQLError(msg, node))

        return None


@cache
def core_implements_one_of_directive() -> bool:
    from graphql.type import directives  # noqa: PLC0415

    return "GraphQLOneOfDirective" in dir(directives)


def validate_one_of_input_object_variable_value(value: dict[str, Any], *, typename: str) -> dict[str, Any]:
    if core_implements_one_of_directive():
        return value

    if len(value) != 1:
        msg = f"OneOf Input Object '{typename}' must specify exactly one key."
        raise GraphQLError(msg)

    field_key, field_value = next(iter(value.items()))

    if field_value is None:
        msg = f"Field '{typename}.{field_key}' must be non-null."
        raise GraphQLError(msg)

    return value


def is_one_of_input_object(input_object_type: GraphQLInputObjectType) -> bool:
    if core_implements_one_of_directive():
        return input_object_type.is_one_of
    return input_object_type.extensions.get("_undine_is_one_of", False)


def get_one_of_input_object_type_extension() -> dict[str, Any]:
    return {"_undine_is_one_of": True}



from __future__ import annotations

import hashlib
from typing import TYPE_CHECKING, Any

from django.core.exceptions import ValidationError

from undine.exceptions import GraphQLErrorGroup, GraphQLRequestParseError, GraphQLValidationError

from .models import PersistedDocument

if TYPE_CHECKING:
    from undine.typing import DjangoRequestProtocol

__all__ = [
    "parse_document_map",
    "register_persisted_documents",
    "to_document_id",
]


def to_document_id(document: str) -> str:
    payload = hashlib.sha256(document.encode("utf-8")).hexdigest()
    return f"sha256:{payload}"


def register_persisted_documents(document_map: dict[str, str]) -> dict[str, str]:
    """
    Register persisted documents.

    :param document_map: Mapping of user defined keys to GraphQL documents to persist.
    :return: Mapping where the keys are the same as given in 'document_map',
             but values have been changed to the 'document_ids' of the persisted documents.
    """
    document_id_map: dict[str, str] = {}
    docs: list[PersistedDocument] = []
    errors: list[GraphQLValidationError] = []

    for key, document in document_map.items():
        document_id = to_document_id(document)
        doc = PersistedDocument(document_id=document_id, document=document)

        try:
            doc.full_clean()
        except ValidationError as err:
            for message in err.messages:
                error = GraphQLValidationError(message, path=["documents", key])
                errors.append(error)
            continue

        docs.append(doc)
        document_id_map[key] = document_id

    if errors:
        raise GraphQLErrorGroup(errors)

    PersistedDocument.objects.bulk_create(
        docs,
        update_conflicts=True,
        update_fields=["document"],
        unique_fields=["document_id"],
    )

    return document_id_map


def parse_document_map(json_data: dict[str, Any]) -> dict[str, str]:
    documents = json_data.get("documents")
    if documents is None:
        msg = "Missing key."
        raise GraphQLRequestParseError(msg, path=["documents"])

    if not isinstance(documents, dict):
        msg = "Value is not a dictionary."
        raise GraphQLRequestParseError(msg, path=["documents"])

    errors: list[GraphQLRequestParseError] = []

    for key, value in documents.items():
        if not isinstance(value, str):
            msg = "Value is not a string."
            error = GraphQLRequestParseError(msg, path=["documents", key])
            errors.append(error)

    if errors:
        raise GraphQLErrorGroup(errors)

    return documents


def default_permission_callback(request: DjangoRequestProtocol, document_map: dict[str, str]) -> None:
    """Default permission callback for persisted documents."""



from __future__ import annotations

from functools import partial
from inspect import cleandoc
from typing import TYPE_CHECKING, Any

from graphql import (
    DirectiveLocation,
    GraphQLArgument,
    GraphQLDirective,
    GraphQLEnumType,
    GraphQLEnumValue,
    GraphQLInputObjectType,
    GraphQLInt,
    GraphQLInterfaceType,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLScalarType,
    GraphQLUnionType,
    specified_directives,
    specified_scalar_types,
)

from undine.exceptions import GraphQLDuplicateTypeError
from undine.utils.registy import Registry

if TYPE_CHECKING:
    from collections.abc import Callable, Collection

    from graphql import (
        GraphQLAbstractType,
        GraphQLField,
        GraphQLInputField,
        GraphQLScalarSerializer,
        GraphQLScalarValueParser,
        GraphQLTypeResolver,
    )
    from graphql.type.definition import GraphQLEnumValueMap, GraphQLInputFieldOutType

    from undine import GQLInfo
    from undine.typing import UniquelyNamedGraphQLElement
    from undine.utils.reflection import FunctionEqualityWrapper


__all__ = [
    "GRAPHQL_REGISTRY",
    "get_or_create_graphql_directive",
    "get_or_create_graphql_enum",
    "get_or_create_graphql_input_object_type",
    "get_or_create_graphql_interface_type",
    "get_or_create_graphql_object_type",
    "get_or_create_graphql_scalar",
    "get_or_create_graphql_union",
    "get_registered_directives",
    "register_builtins",
]


GRAPHQL_REGISTRY: Registry[str, UniquelyNamedGraphQLElement] = Registry()
"""
Caches created GraphQL elements by their names so that they can be reused during schema creation
since a GraphQL Schema cannot contain multiple elements with the same name.
"""


def get_or_create_graphql_object_type(
    *,
    name: str,
    fields: dict[str, GraphQLField] | FunctionEqualityWrapper[dict[str, GraphQLField]],
    interfaces: Collection[GraphQLInterfaceType] | None = None,
    description: str | None = None,
    is_type_of: Callable[[Any, GQLInfo], bool] | None = None,
    extensions: dict[str, Any] | None = None,
) -> GraphQLObjectType:
    """
    Either create a new 'GraphQLObjectType' or get an existing one with the same name.
    Checks that the existing element is indeed a 'GraphQLObjectType', and raises an error
    if it is not the same.
    """
    object_type = GraphQLObjectType(
        name=name,
        fields=fields,
        interfaces=interfaces,
        description=description,
        is_type_of=is_type_of,  # type: ignore[arg-type]
        extensions=extensions,
    )

    if name in GRAPHQL_REGISTRY:
        existing = GRAPHQL_REGISTRY[name]

        if (
            not isinstance(existing, GraphQLObjectType)
            or existing._fields != object_type._fields
            or existing._interfaces != object_type._interfaces  # noqa: SLF001
            or existing.is_type_of != object_type.is_type_of
            or existing.description != object_type.description
            or existing.extensions != object_type.extensions
        ):
            raise GraphQLDuplicateTypeError(
                name=name,
                type_new=object_type,
                type_existing=existing,
            )

        return existing

    GRAPHQL_REGISTRY[name] = object_type
    return object_type


def get_or_create_graphql_input_object_type(
    *,
    name: str,
    fields: dict[str, GraphQLInputField] | FunctionEqualityWrapper[dict[str, GraphQLInputField]],
    description: str | None = None,
    extensions: dict[str, Any] | None = None,
    out_type: GraphQLInputFieldOutType | None = None,
    is_one_of: bool = False,
) -> GraphQLInputObjectType:
    """
    Either create a new 'GraphQLInputObjectType' or get an existing one with the same name.
    Checks that the existing element is indeed a 'GraphQLInputObjectType', and raises an error
    if it is not the same.
    """
    from undine.utils.graphql.validation_rules.one_of_input_object import (  # noqa: PLC0415
        core_implements_one_of_directive,
        get_one_of_input_object_type_extension,
        validate_one_of_input_object_variable_value,
    )

    if core_implements_one_of_directive():
        input_object_type = GraphQLInputObjectType(
            name=name,
            fields=fields,
            description=description,
            extensions=extensions,
            out_type=out_type,
            is_one_of=is_one_of,
        )

    else:
        input_object_type = GraphQLInputObjectType(
            name=name,
            fields=fields,
            description=description,
            extensions=(extensions or {}) | get_one_of_input_object_type_extension(),
            out_type=partial(validate_one_of_input_object_variable_value, typename=name),
        )

    if name in GRAPHQL_REGISTRY:
        existing = GRAPHQL_REGISTRY[name]

        if (
            not isinstance(existing, GraphQLInputObjectType)
            or existing._fields != input_object_type._fields
            or existing.description != input_object_type.description
            or existing.extensions != input_object_type.extensions
            or existing.out_type != input_object_type.out_type
            or getattr(existing, "is_one_of", False) != getattr(existing, "is_one_of", False)
        ):
            raise GraphQLDuplicateTypeError(
                name=name,
                type_new=input_object_type,
                type_existing=existing,
            )

        return existing

    GRAPHQL_REGISTRY[name] = input_object_type
    return input_object_type


def get_or_create_graphql_interface_type(
    *,
    name: str,
    fields: dict[str, GraphQLField] | FunctionEqualityWrapper[dict[str, GraphQLField]],
    interfaces: Collection[GraphQLInterfaceType] | None = None,
    resolve_type: GraphQLTypeResolver | None = None,
    description: str | None = None,
    extensions: dict[str, Any] | None = None,
) -> GraphQLInterfaceType:
    """
    Either create a new 'GraphQLInterfaceType' or get an existing one with the same name.
    Checks that the existing element is indeed a 'GraphQLInterfaceType', and raises an error
    if it is not the same.
    """
    interface_type = GraphQLInterfaceType(
        name=name,
        fields=fields,
        interfaces=interfaces,
        resolve_type=resolve_type,
        description=description,
        extensions=extensions,
    )

    if name in GRAPHQL_REGISTRY:
        existing = GRAPHQL_REGISTRY[name]

        if (
            not isinstance(existing, GraphQLInterfaceType)
            or existing._fields != interface_type._fields
            or existing._interfaces != interface_type._interfaces  # noqa: SLF001
            or existing.resolve_type != interface_type.resolve_type
            or existing.description != interface_type.description
            or existing.extensions != interface_type.extensions
        ):
            raise GraphQLDuplicateTypeError(
                name=name,
                type_new=interface_type,
                type_existing=existing,
            )

        return existing

    GRAPHQL_REGISTRY[name] = interface_type
    return interface_type


def get_or_create_graphql_enum(
    *,
    name: str,
    values: GraphQLEnumValueMap | dict[str, Any],
    description: str | None = None,
    extensions: dict[str, Any] | None = None,
) -> GraphQLEnumType:
    """
    Either create a new 'GraphQLEnumType' or get an existing one with the same name.
    Checks that the existing element is indeed a 'GraphQLEnumType', and raises an error
    if it is not the same.
    """
    for key, value in values.items():
        if isinstance(value, str):
            values[key] = GraphQLEnumValue(value=key, description=value)

    enum = GraphQLEnumType(
        name=name,
        values=values,
        description=description,
        extensions=extensions,
    )

    if name in GRAPHQL_REGISTRY:
        existing = GRAPHQL_REGISTRY[name]

        if (
            not isinstance(existing, GraphQLEnumType)
            or existing.values != enum.values
            or existing.description != enum.description
            or existing.extensions != enum.extensions
        ):
            raise GraphQLDuplicateTypeError(
                name=name,
                type_new=enum,
                type_existing=existing,
            )

        return existing

    GRAPHQL_REGISTRY[name] = enum
    return enum


def get_or_create_graphql_union(
    *,
    name: str,
    types: list[GraphQLObjectType] | FunctionEqualityWrapper[list[GraphQLObjectType]],
    resolve_type: Callable[[Any, GQLInfo, GraphQLAbstractType], str | None] | None = None,
    description: str | None = None,
    extensions: dict[str, Any] | None = None,
) -> GraphQLUnionType:
    """
    Either create a new 'GraphQLUnionType' or get an existing one with the same name.
    Checks that the existing element is indeed a GraphQLUnionType, and raises an error
    if it is not the same.
    """
    union = GraphQLUnionType(
        name=name,
        types=types,
        resolve_type=resolve_type,  # type: ignore[arg-type]
        description=description,
        extensions=extensions,
    )

    if name in GRAPHQL_REGISTRY:
        existing = GRAPHQL_REGISTRY[name]

        if (
            not isinstance(existing, GraphQLUnionType)
            or existing._types != union._types  # noqa: SLF001
            or existing.resolve_type != union.resolve_type
            or existing.description != union.description
            or existing.extensions != union.extensions
        ):
            raise GraphQLDuplicateTypeError(
                name=name,
                type_new=union,
                type_existing=existing,
            )

        return existing

    GRAPHQL_REGISTRY[name] = union
    return union


def get_or_create_graphql_scalar(
    *,
    name: str,
    serialize: GraphQLScalarSerializer | None = None,
    parse_value: GraphQLScalarValueParser | None = None,
    description: str | None = None,
    specified_by_url: str | None = None,
    extensions: dict[str, Any] | None = None,
) -> GraphQLScalarType:
    """
    Either create a new 'GraphQLScalarType' or get an existing one with the same name.
    Checks that the existing element is indeed a 'GraphQLScalarType', and raises an error
    if it is not the same.
    """
    scalar = GraphQLScalarType(
        name=name,
        serialize=serialize,
        parse_value=parse_value,
        description=description,
        specified_by_url=specified_by_url,
        extensions=extensions,
    )

    if name in GRAPHQL_REGISTRY:
        existing = GRAPHQL_REGISTRY[name]

        if (
            not isinstance(existing, GraphQLScalarType)
            or existing.serialize != scalar.serialize
            or existing.parse_value != scalar.parse_value
            or existing.description != scalar.description
            or existing.specified_by_url != scalar.specified_by_url
            or existing.extensions != scalar.extensions
        ):
            raise GraphQLDuplicateTypeError(
                name=name,
                type_new=scalar,
                type_existing=existing,
            )

        return existing

    GRAPHQL_REGISTRY[name] = scalar
    return scalar


def get_or_create_graphql_directive(
    *,
    name: str,
    locations: list[DirectiveLocation],
    args: dict[str, GraphQLArgument] | None = None,
    is_repeatable: bool = False,
    description: str | None = None,
    extensions: dict[str, Any] | None = None,
) -> GraphQLDirective:
    """
    Either create a new 'GraphQLDirective' or get an existing one with the same name.
    Checks that the existing element is indeed a 'GraphQLDirective', and raises an error
    if it is not the same.
    """
    directive = GraphQLDirective(
        name=name,
        locations=locations,
        args=args,
        is_repeatable=is_repeatable,
        description=description,
        extensions=extensions,
    )

    if name in GRAPHQL_REGISTRY:
        existing = GRAPHQL_REGISTRY[name]

        if not isinstance(existing, GraphQLDirective) or directive != existing:
            raise GraphQLDuplicateTypeError(
                name=name,
                type_new=directive,
                type_existing=existing,
            )

        return existing

    GRAPHQL_REGISTRY[name] = directive
    return directive


GraphQLComplexityDirective = GraphQLDirective(
    name="complexity",
    locations=[DirectiveLocation.FIELD_DEFINITION],
    args={
        "value": GraphQLArgument(
            GraphQLNonNull(GraphQLInt),
            out_name="value",
        ),
    },
    description=cleandoc(
        """
        Indicate the complexity of resolving a field, counted towards
        the maximum query complexity of resolving a root type field.
        """
    ),
)
"""Used to indicate the complexity of resolving a given field. See `Field.complexity`."""


GraphQLOneOfDirective = GraphQLDirective(
    name="oneOf",
    locations=[DirectiveLocation.INPUT_OBJECT],
    description="Indicates exactly one field must be supplied and this field must not be `null`.",
)
"""Used to indicate an Input Object is a OneOf Input Object."""


GraphQLAtomicDirective = GraphQLDirective(
    name="atomic",
    locations=[DirectiveLocation.MUTATION],
    description="Indicates that all mutations in the operation should be executed atomically.",
)
"""Used to indicate that all mutations in the operation should be executed atomically."""


def register_builtins() -> None:
    from undine.utils.graphql.validation_rules import core_implements_one_of_directive  # noqa: PLC0415

    for name, scalar in specified_scalar_types.items():
        GRAPHQL_REGISTRY[name] = scalar

    for directive in specified_directives:
        GRAPHQL_REGISTRY[directive.name] = directive

    GRAPHQL_REGISTRY[GraphQLComplexityDirective.name] = GraphQLComplexityDirective
    GRAPHQL_REGISTRY[GraphQLAtomicDirective.name] = GraphQLAtomicDirective

    # graphql-core < 3.3.0
    if not core_implements_one_of_directive():
        GRAPHQL_REGISTRY[GraphQLOneOfDirective.name] = GraphQLOneOfDirective


def get_registered_directives() -> tuple[GraphQLDirective, ...]:
    return tuple(value for value in GRAPHQL_REGISTRY.values() if isinstance(value, GraphQLDirective))



from __future__ import annotations

import json
from abc import ABC, abstractmethod
from contextlib import suppress
from typing import TYPE_CHECKING, Any

import pytest
from django.contrib.auth import get_user_model
from django.test.client import MULTIPART_CONTENT, AsyncClient, Client
from graphql import FormattedExecutionResult

from undine.http.files import extract_files
from undine.persisted_documents.utils import to_document_id
from undine.settings import undine_settings
from undine.typing import WebSocketASGIScope

from .query_logging import capture_database_queries

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator
    from http.cookies import SimpleCookie

    from django.contrib.auth.models import User
    from django.core.files import File
    from graphql import GraphQLFormattedError

    from undine.typing import DjangoTestClientResponseProtocol

    from .query_logging import DBQueryData
    from .websocket import TestWebSocket

with suppress(ImportError):
    from .websocket import TestWebSocket


__all__ = [
    "AsyncGraphQLClient",
    "GraphQLClient",
    "GraphQLClientHTTPResponse",
    "GraphQLClientWebSocketResponse",
]


IS_CHANNELS_INSTALLED: bool = "TestWebSocket" in globals()


class WebSocketMixin:
    """Mixin to support GraphQL over WebSocket requests."""

    cookies: SimpleCookie

    async def over_websocket(
        self,
        document: str,
        *,
        variables: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
        operation_name: str | None = None,
        use_persisted_document: bool = False,
        count_queries: bool = False,
    ) -> AsyncGenerator[GraphQLClientWebSocketResponse, None]:
        """
        Send a GraphQL over WebSocket request and yield the execution results.

        :param document: GraphQL document string to send in the request.
        :param variables: GraphQL variables for the document.
        :param headers: Headers for the request.
        :param operation_name: If given document includes multiple operations,
                               this is required to select the operation to execute.
        :param use_persisted_document: Instead of sending the whole document,
                                       convert it to a persisted document ID and send that.
        :param count_queries: If True, count the number of queries executed during the request.
        """
        variables = variables or {}
        body: dict[str, Any] = {}

        if use_persisted_document:
            body["documentId"] = to_document_id(document)
        else:
            body["query"] = document

        if variables:
            body["variables"] = variables
        if operation_name is not None:
            body["operationName"] = operation_name

        websocket = self.websocket()

        if headers is not None:
            websocket.scope["headers"] += [
                (name.encode("ascii"), value.encode("ascii")) for name, value in headers.items()
            ]

        async with websocket:
            await websocket.connection_init()

            with capture_database_queries(enabled=count_queries) as queries:
                initial_result = await websocket.subscribe(body)

            if initial_result["type"] == "error":
                result = FormattedExecutionResult(data=None, errors=initial_result["payload"])
                yield GraphQLClientWebSocketResponse(result=result, database_queries=queries)
                return

            yield GraphQLClientWebSocketResponse(result=initial_result["payload"], database_queries=queries)

            while True:
                with capture_database_queries(enabled=count_queries) as queries:
                    message = await websocket.receive()

                match message["type"]:
                    case "complete":
                        return

                    case "next":
                        yield GraphQLClientWebSocketResponse(result=message["payload"], database_queries=queries)
                        continue

                    case "error":
                        result = FormattedExecutionResult(data=None, errors=message["payload"])
                        yield GraphQLClientWebSocketResponse(result=result, database_queries=queries)
                        return

                    case _:  # pragma: no cover
                        msg = f"Unexpected message type: {message['type']}"
                        raise RuntimeError(msg)

    def websocket(self, **kwargs: Any) -> TestWebSocket:
        """Create a testing websocket."""
        if not IS_CHANNELS_INSTALLED:  # pragma: no cover
            msg = "The `channels` library is not installed. Cannot create websocket."
            raise RuntimeError(msg)

        scope = self._get_default_scope()
        scope.update(kwargs)  # type: ignore[typeddict-item]

        return TestWebSocket(scope=scope)

    def _get_default_scope(self) -> WebSocketASGIScope:
        # From 'django.test.client.AsyncRequestFactory._base_scope'
        cookies = (f"{morsel.key}={morsel.coded_value}".encode("ascii") for morsel in self.cookies.values())
        path = f"/{undine_settings.WEBSOCKET_PATH}"
        return WebSocketASGIScope(  # type: ignore[typeddict-item]
            type="websocket",
            asgi={"version": "3.0"},
            http_version="1.1",
            scheme="ws",
            server=("testserver", 80),
            client=("127.0.0.1", 0),
            root_path="",
            path=path,
            raw_path=path.encode("ascii"),
            query_string=b"",
            headers=[
                (b"cookie", b"; ".join(sorted(cookies))),
                (b"host", b"testserver"),
                (b"connection", b"Upgrade"),
                (b"upgrade", b"websocket"),
                (b"sec-websocket-version", b"13"),
                (b"sec-websocket-key", b"RKr31GF3kXZqsXjVT7s3Mg=="),
                (b"sec-websocket-protocol", b"graphql-transport-ws"),
            ],
            subprotocols=["graphql-transport-ws"],
            state={},
            extensions={"websocket.http.response": {}},
            cookies={k: str(v) for k, v in self.cookies.items()},
            path_remaining="",
            url_route={"args": (), "kwargs": {}},
            #
            # 'user' and 'session' are set in the `WebSocketContextManager` `AuthMiddlewareStack`.
        )  # type: ignore[typeddict-item]


class GraphQLClient(WebSocketMixin, Client):
    """A GraphQL client for testing."""

    def __call__(
        self,
        document: str,
        *,
        variables: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
        operation_name: str | None = None,
        use_persisted_document: bool = False,
        count_queries: bool = False,
    ) -> GraphQLClientHTTPResponse:
        """
        Execute a GraphQL operation synchronously.

        :param document: GraphQL document string to send in the request.
        :param variables: GraphQL variables for the document.
        :param headers: Headers for the request.
        :param operation_name: If given document includes multiple operations,
                               this is required to select the operation to execute.
        :param use_persisted_document: Instead of sending the whole document,
                                       convert it to a persisted document ID and send that.
        :param count_queries: If True, count the number of queries executed during the request.
        """
        variables = variables or {}
        body: dict[str, Any] = {}

        if use_persisted_document:
            body["documentId"] = to_document_id(document)
        else:
            body["query"] = document

        if variables:
            body["variables"] = variables
        if operation_name is not None:
            body["operationName"] = operation_name

        files = extract_files(variables)
        data = _create_multipart_data(body, files) if files else body
        content_type = MULTIPART_CONTENT if files else "application/json"

        with capture_database_queries(enabled=count_queries) as results:
            response: DjangoTestClientResponseProtocol = self.post(  # type: ignore[assignment]
                path=f"/{undine_settings.GRAPHQL_PATH}",
                data=data,
                content_type=content_type,
                headers=headers,
            )

        return GraphQLClientHTTPResponse(response, results)

    def login_with_superuser(self, username: str = "admin", **kwargs: Any) -> User:
        """Create a superuser and log in as that user."""
        defaults = {
            "is_staff": True,
            "is_superuser": True,
            "email": "superuser@django.com",
            **kwargs,
        }
        user, _ = get_user_model().objects.get_or_create(username=username, defaults=defaults)
        self.force_login(user)
        return user

    def login_with_regular_user(self, username: str = "user", **kwargs: Any) -> User:
        """Create a regular user and log in as that user."""
        defaults = {
            "is_staff": False,
            "is_superuser": False,
            "email": "user@django.com",
            **kwargs,
        }
        user, _ = get_user_model().objects.get_or_create(username=username, defaults=defaults)
        self.force_login(user)
        return user


class AsyncGraphQLClient(WebSocketMixin, AsyncClient):
    """An async GraphQL client for testing."""

    async def __call__(
        self,
        document: str,
        *,
        variables: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
        operation_name: str | None = None,
        use_persisted_document: bool = False,
        count_queries: bool = False,
    ) -> GraphQLClientHTTPResponse:
        """
        Execute a GraphQL operation asynchronously.

        :param document: GraphQL document string to send in the request.
        :param variables: GraphQL variables for the document.
        :param headers: Headers for the request.
        :param operation_name: If given document includes multiple operations,
                               this is required to select the operation to execute.
        :param use_persisted_document: Instead of sending the whole document,
                                       convert it to a persisted document ID and send that.
        :param count_queries: If True, count the number of queries executed during the request.
        """
        variables = variables or {}
        body: dict[str, Any] = {}

        if use_persisted_document:
            body["documentId"] = to_document_id(document)
        else:
            body["query"] = document

        if variables:
            body["variables"] = variables
        if operation_name is not None:
            body["operationName"] = operation_name

        files = extract_files(variables)
        data = _create_multipart_data(body, files) if files else body
        content_type = MULTIPART_CONTENT if files else "application/json"

        with capture_database_queries(enabled=count_queries) as results:
            response: DjangoTestClientResponseProtocol = await self.post(  # type: ignore[assignment]
                path=f"/{undine_settings.GRAPHQL_PATH}",
                data=data,
                content_type=content_type,
                headers=headers,
            )

        return GraphQLClientHTTPResponse(response, results)

    async def login_with_superuser(self, username: str = "admin", **kwargs: Any) -> User:
        """Create a superuser and log in as that user."""
        defaults = {
            "is_staff": True,
            "is_superuser": True,
            "email": "superuser@django.com",
            **kwargs,
        }
        user, _ = await get_user_model().objects.aget_or_create(username=username, defaults=defaults)
        self.force_login(user)
        return user

    async def login_with_regular_user(self, username: str = "user", **kwargs: Any) -> User:
        """Create a regular user and log in as that user."""
        defaults = {
            "is_staff": False,
            "is_superuser": False,
            "email": "user@django.com",
            **kwargs,
        }
        user, _ = await get_user_model().objects.aget_or_create(username=username, defaults=defaults)
        self.force_login(user)
        return user


class BaseGraphQLClientResponse(ABC):
    """Base class for GraphQLClient responses."""

    DB_QUERIES_NOT_ENABLED_MSG = (
        "Database queries have not been captured. "
        "Enable them using the `count_queries` parameter when calling the client."
    )

    def __init__(self, database_queries: DBQueryData) -> None:
        self._database_queries = database_queries

    @property
    @abstractmethod
    def json(self) -> FormattedExecutionResult:
        """Return the JSON content of the response."""

    # Concrete methods

    def __str__(self) -> str:
        return json.dumps(self.json, indent=2, sort_keys=True, default=str)

    def __repr__(self) -> str:
        return repr(self.json)

    def __getitem__(self, item: str) -> Any:
        return self.json[item]  # type: ignore[literal-required]

    def __contains__(self, item: str) -> bool:
        return item in self.json

    @property
    def queries(self) -> list[str]:
        """Return a list of the database queries that were executed."""
        if not self._database_queries.enabled:
            return pytest.fail(self.DB_QUERIES_NOT_ENABLED_MSG)

        return [info.sql for info in self._database_queries.queries]

    @property
    def query_count(self) -> int:
        """Return the number of database queries that were executed."""
        if not self._database_queries.enabled:
            return pytest.fail(self.DB_QUERIES_NOT_ENABLED_MSG)

        return self._database_queries.count

    @property
    def query_log(self) -> str:
        """Return a string representation of the database queries that were executed."""
        if not self._database_queries.enabled:
            return pytest.fail(self.DB_QUERIES_NOT_ENABLED_MSG)

        return self._database_queries.log

    @property
    def data(self) -> dict[str, Any] | None:
        """Return the data from the response content."""
        return self.json["data"]

    @property
    def results(self) -> dict[str, Any] | list[Any] | None:
        """
        Return the results from the first top-level field in the response content.

        >>> self.json
        {"data": {"foo": {"name": "bar"}}}
        >>> self.results
        {"name": "bar"}
        """
        data = self.data or {}
        try:
            return next(iter(data.values()))
        except StopIteration:
            msg = f"No query object not found in response content\nContent: {self.json}"
            return pytest.fail(msg)

    @property
    def edges(self) -> list[dict[str, Any]]:
        """
        Return edges from the first top-level field in the response content.

        >>> self.json
        {"data": {"foo": {"edges": [{"node": {"name": "bar"}}]}}}
        >>> self.edges
        [{"node": {"name": "bar"}}]
        """
        results = self.results
        if not isinstance(results, dict):
            msg = f"Edges not found in response content\nContent: {self.json}"
            return pytest.fail(msg)

        try:
            return results["edges"]
        except (KeyError, TypeError):
            msg = f"Edges not found in response content\nContent: {self.json}"
            return pytest.fail(msg)

    def node(self, index: int) -> dict[str, Any]:
        """
        Return the node at the given index in the response content edges.

        >>> self.json
        {"data": {"foo": {"edges": [{"node": {"name": "bar"}}]}}}
        >>> self.node(0)
        {"name": "bar"}
        """
        try:
            return self.edges[index]["node"]
        except (IndexError, TypeError):
            msg = f"Node {index!r} not found in response content\nContent: {self.json}"
            return pytest.fail(msg)

    @property
    def has_errors(self) -> bool:
        """Are there any errors in the response?"""
        return "errors" in self.json and self.json.get("errors") is not None

    @property
    def errors(self) -> list[GraphQLFormattedError]:
        """
        Return all errors.

        >>> self.json
        {"errors": [{"message": "bar", "path": [...], ...}]}
        >>> self.errors
        [{"message": "bar", "path": [...], ...}]
        """
        try:
            return self.json["errors"]
        except (KeyError, TypeError):
            msg = f"Errors not found in response content\nContent: {self.json}"
            return pytest.fail(msg)

    def error_message(self, selector: int | str) -> str:
        """
        Return the error message from the errors list...

        1) in the given index

        >>> self.json
        {"errors": [{"message": "bar", ...}]}
        >>> self.error_message(0)
        "bar"

        2) in the given path:

        >>> self.json
        {"errors": [{"message": "bar", "path": ["fizz", "buzz", "foo"], ...}]}
        >>> self.error_message("foo")
        "bar"
        """
        if isinstance(selector, int):
            try:
                return self.errors[selector]["message"]
            except (IndexError, KeyError, TypeError):
                msg = f"Errors message not found from index {selector}\nContent: {self.json}"
                return pytest.fail(msg)
        else:
            try:
                return next(error["message"] for error in self.errors if error["path"][-1] == selector)
            except (StopIteration, KeyError, TypeError):
                msg = f"Errors message not found from path {selector!r}\nContent: {self.json}"
                return pytest.fail(msg)

    def assert_query_count(self, count: int) -> None:
        if not self._database_queries.enabled:
            pytest.fail(self.DB_QUERIES_NOT_ENABLED_MSG)

        if self.query_count != count:
            msg = f"Expected {count} queries, got {self.query_count}.\n{self.query_log}"
            pytest.fail(msg)


class GraphQLClientHTTPResponse(BaseGraphQLClientResponse):
    """An HTTP response from the GraphQLClient."""

    def __init__(self, response: DjangoTestClientResponseProtocol, database_queries: DBQueryData) -> None:
        self.response = response
        super().__init__(database_queries=database_queries)

    @property
    def json(self) -> FormattedExecutionResult:
        """Return the JSON content of the response."""
        return self.response.json()  # type: ignore[return-value]

    @property
    def status_code(self) -> int:
        """Return the status code of the response."""
        return self.response.status_code


class GraphQLClientWebSocketResponse(BaseGraphQLClientResponse):
    """A WebSocket response from the GraphQLClient."""

    def __init__(self, result: FormattedExecutionResult, database_queries: DBQueryData) -> None:
        self.result = result
        super().__init__(database_queries=database_queries)

    @property
    def json(self) -> FormattedExecutionResult:
        """Return the JSON content of the response."""
        return self.result


def _create_multipart_data(body: dict[str, Any], files: dict[File, list[str]]) -> dict[str, Any]:
    path_map: dict[str, list[str]] = {}
    files_map: dict[str, File] = {}
    for i, (file, path) in enumerate(files.items()):
        path_map[str(i)] = path
        files_map[str(i)] = file

    return {
        "operations": json.dumps(body),
        "map": json.dumps(path_map),
        **files_map,
    }



from __future__ import annotations

from typing import TYPE_CHECKING, Generic, TypeVar

from undine.exceptions import RegistryDuplicateError, RegistryMissingTypeError
from undine.utils.reflection import get_instance_name

if TYPE_CHECKING:
    from collections.abc import ItemsView, Iterator, KeysView, ValuesView

__all__ = [
    "Registry",
]


From = TypeVar("From")
To = TypeVar("To")


class Registry(Generic[From, To]):
    """
    A registry for values that need to be globally available.
    Verifies that a value for a given key is only registered once.
    """

    def __init__(self) -> None:
        self.__registry: dict[From, To] = {}
        self.__name = get_instance_name()

    def __getitem__(self, key: From) -> To:
        try:
            return self.__registry[key]
        except KeyError as error:
            raise RegistryMissingTypeError(registry_name=self.__name, key=key) from error

    def __setitem__(self, key: From, value: To) -> None:
        if key in self.__registry:
            raise RegistryDuplicateError(key=key, value=self.__registry[key], registry_name=self.__name)
        self.__registry[key] = value

    def __contains__(self, key: From) -> bool:
        return key in self.__registry

    def __iter__(self) -> Iterator[From]:
        return iter(self.__registry)

    def keys(self) -> KeysView[From]:
        return self.__registry.keys()

    def values(self) -> ValuesView[To]:
        return self.__registry.values()

    def items(self) -> ItemsView[From, To]:
        return self.__registry.items()

    def clear(self) -> None:
        self.__registry.clear()
